/* 

   test_diagram_isotopic.c : Unit tests for the code in pd_diagram_isotopics.c.


*/

#ifdef HAVE_CONFIG_H
#include"config.h"
#endif

#ifdef HAVE_STDIO_H
#include<stdio.h>
#endif  

#ifdef HAVE_STRING_H
#include<string.h>
#endif

#ifdef HAVE_STDINT_H
#include<stdint.h>
#endif

#ifdef HAVE_STDLIB_H
#include<stdlib.h>
#endif

#ifdef HAVE_GSL_GSL_PERMUTATION_H
#include<gsl/gsl_permutation.h>
#endif

#ifdef HAVE_STDBOOL_H
#include<stdbool.h>
#endif

#ifdef HAVE_ASSERT_H
#include<assert.h>
#endif

#ifdef HAVE_ARGTABLE2_H
#include<argtable2.h> 
#endif

#include<plcTopology.h>

#include<pd_multidx.h>
#include<pd_perm.h>
#include<pd_dihedral.h>
#include<pd_cyclic.h>

#include<pd_isomorphisms.h>

int PD_VERBOSE=50;

/* Idea: Given up a diagram, we'll generate, systematically, all the
   renumberings which should result in isomorphic diagrams: we should 
   be able to then see if pd_diagram_isotopic picks them up. */

bool all_diagram_isotopies_test(pd_code_t *pd,char *name)
{

  printf("all diagram isotopies test for pd code %s\n",name);
  printf("a %d-crossing, %d-component pd code.\n",pd->ncross,pd->ncomps);
  printf("--------------------------------------------\n");

  printf("testing pd_ok...");

  if (!pd_ok(pd)) { 

    pd_printf("fail (pd %PD does not pass pd_ok)\n",pd);
    return false;

  }

  printf("pass (pd_ok).\n");

  printf("constructing multidx for iteration...");

  pd_idx_t **compsizes;
  pd_idx_t i;

  compsizes = calloc(pd->ncomps,sizeof(pd_idx_t *));
  assert(compsizes != NULL);
  for(i=0;i<pd->ncomps;i++) { compsizes[i] = &(pd->comp[i].nedges); }

  pd_multidx_t *idx = pd_new_multidx(pd->ncomps,(void **)(compsizes),cyclic_ops);

  printf("pass (%d possible variations of this pd_code)\n",pd_multidx_nvals(idx));

  for(i=0;i<pd_multidx_nvals(idx);i++,pd_increment_multidx(idx)) { 

    pd_code_t *pdCopy = pd_copy(pd);
    pd_idx_t j;

    for(j=0;j<pd->ncomps;j++) { 

      pd_cyclic_t *d = (pd_cyclic_t *)(idx->obj[j]);
      pd_renumber_component(pdCopy,j,d->map[0]);

    }

    if (!pd_diagram_isotopic(pd,pdCopy)) {

      pd_printf("fail at %MULTIDX\n",pd,idx);
      
      for(j=0;j<pd->ncomps;j++) { 

	pd_printf("\t %CYCLIC \n",pd,(pd_cyclic_t *)(idx->obj[j]));

      }

      pd_printf("renumbered and reoriented pd \n"
		"-----------------------------\n"
		"%PD\n",pdCopy);
      pd_printf("-----------------------------\n"
		"is not diagram-isotopic to original pd\n"
		"-----------------------------\n"
		"%PD\n",pd);
      printf("-----------------------------\n");

      return false;

    }

    pd_code_free(&pdCopy);

  }

  pd_free_multidx(&idx); 
  printf("pass (%d of %d reorder/renumberings detected as diagram-isotopic)\n",i,i);

  return true;
}

/* machine-generated diagram-isomorphism test c code */

pd_code_t *pd_create_3_0() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 3;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_1() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 2;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 3;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 0;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_2() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_3() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_4() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_5() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 2;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_3_6() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(3);
assert(pd != NULL);
pd->ncross = 3;
pd->nedges = 6;
pd->ncomps = 1;
pd->nfaces = 5;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

bool test_diagrams() {
pd_code_t *pd;
printf("testing diagram pd_create_3_0...");
pd = pd_create_3_0();
if (!all_diagram_isotopies_test(pd,"pd_create_3_0")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_1...");
pd = pd_create_3_1();
if (!all_diagram_isotopies_test(pd,"pd_create_3_1")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_2...");
pd = pd_create_3_2();
if (!all_diagram_isotopies_test(pd,"pd_create_3_2")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_3...");
pd = pd_create_3_3();
if (!all_diagram_isotopies_test(pd,"pd_create_3_3")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_4...");
pd = pd_create_3_4();
if (!all_diagram_isotopies_test(pd,"pd_create_3_4")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_5...");
pd = pd_create_3_5();
if (!all_diagram_isotopies_test(pd,"pd_create_3_5")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_3_6...");
pd = pd_create_3_6();
if (!all_diagram_isotopies_test(pd,"pd_create_3_6")) {
  return false;
}
pd_code_free(&pd);

return true;
}



/* machine-generated diagram-isomorphism test c code */

pd_code_t *pd_create_6_0() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 3;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 5;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 3;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 4;
pd->face[6].orient[1] = 1;

pd->face[6].edge[2] = 8;
pd->face[6].orient[2] = 0;

pd->face[7].nedges = 3;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 0;

pd->face[7].edge[2] = 7;
pd->face[7].orient[2] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_1() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 4;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 0;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 10;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_2() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 11;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 6;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_3() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 6;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 10;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_4() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 4;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_5() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 6;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 6;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_6() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 6;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_7() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 4;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;

pd->comp[3].nedges = 2;
pd->comp[3].tag = 'D';

pd->comp[3].edge = calloc(pd->comp[3].nedges,sizeof(pd_idx_t));
assert(pd->comp[3].edge != NULL);

pd->comp[3].edge[0] = 10;
pd->comp[3].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_8() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_9() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 9;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 4;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_10() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 6;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_11() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 2;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 11;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_12() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 4;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 2;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 6;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_13() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_14() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 8;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_15() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 9;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 5;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_16() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 6;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 6;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_17() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[5].edge[2] = 7;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 0;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_18() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_19() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_20() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_21() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_22() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_23() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 6;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 3;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 4;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_24() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 6;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 3;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 4;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_25() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_26() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 0;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_27() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 9;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_28() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 4;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_29() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 4;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_30() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 4;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_31() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 4;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_32() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_33() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 2;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 3;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_34() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_35() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_36() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 3;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_37() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_38() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 4;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[4].edge[3] = 11;
pd->face[4].orient[3] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_39() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_40() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 8;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_41() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_42() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_43() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_44() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 9;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_45() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 9;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_46() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 7;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_47() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_48() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 4;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_49() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_50() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_51() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_52() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_53() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_54() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_55() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_56() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_57() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_58() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_59() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 1;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_60() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_61() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_62() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_63() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_64() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_65() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_66() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_67() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_68() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_69() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_70() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 9;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_71() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_72() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_73() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_74() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 9;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_75() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_76() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_77() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 2;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 3;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_78() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 2;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 3;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_79() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_80() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_81() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_82() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_83() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_84() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_85() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_86() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_87() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 5;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 0;

pd->face[3].edge[4] = 6;
pd->face[3].orient[4] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_88() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 8;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 5;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 0;

pd->face[3].edge[4] = 4;
pd->face[3].orient[4] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_89() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 5;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 1;

pd->face[3].edge[4] = 7;
pd->face[3].orient[4] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_90() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 5;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 5;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 1;

pd->face[3].edge[4] = 7;
pd->face[3].orient[4] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_91() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 9;
pd->face[5].orient[2] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 5;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_92() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 3;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[5].edge[2] = 5;
pd->face[5].orient[2] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 0;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_93() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 6;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 7;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_94() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 8;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_95() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_96() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_97() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_98() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 4;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_99() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_100() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_101() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 0;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_102() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_103() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 4;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;

pd->comp[3].nedges = 2;
pd->comp[3].tag = 'D';

pd->comp[3].edge = calloc(pd->comp[3].nedges,sizeof(pd_idx_t));
assert(pd->comp[3].edge != NULL);

pd->comp[3].edge[0] = 10;
pd->comp[3].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 4;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_104() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_105() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_106() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_107() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_108() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 5;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_109() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_110() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_111() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_112() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_113() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_114() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_115() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_116() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_117() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_118() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_119() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_120() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 3;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 1;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_121() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_122() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_123() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_124() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 9;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_125() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_126() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_127() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 4;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_128() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_129() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_130() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_131() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_132() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_133() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_134() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_135() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_136() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_137() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_138() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_139() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_140() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 6;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_141() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 4;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_142() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_143() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_144() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_145() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 2;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_146() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_147() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_148() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_149() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_150() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_151() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_152() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_153() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_154() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_155() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_156() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_157() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_158() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 6;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_159() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_160() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_161() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_162() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_163() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_164() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_165() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 1;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_166() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 1;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_167() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_168() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_169() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 2;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_170() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_171() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_172() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_173() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_174() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_175() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_176() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_177() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_178() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_179() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_180() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_181() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_182() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_183() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_184() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 7;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_185() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_186() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_187() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_188() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 8;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_189() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_190() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_191() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_192() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_193() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_194() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_195() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_196() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_197() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_198() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_199() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_200() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 8;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_201() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 1;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_202() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_203() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_204() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 8;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_205() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 1;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_206() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 1;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_207() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_208() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 11;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_209() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 4;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;

pd->comp[3].nedges = 2;
pd->comp[3].tag = 'D';

pd->comp[3].edge = calloc(pd->comp[3].nedges,sizeof(pd_idx_t));
assert(pd->comp[3].edge != NULL);

pd->comp[3].edge[0] = 10;
pd->comp[3].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_210() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_211() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_212() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_213() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_214() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 2;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 3;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 4;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 1;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 1;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 2;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_215() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_216() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_217() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_218() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_219() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_220() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_221() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_222() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_223() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 3;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_224() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_225() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_226() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_227() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_228() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_229() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_230() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_231() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_232() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_233() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_234() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 1;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_235() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_236() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_237() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_238() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_239() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_240() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_241() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_242() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 8;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_243() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 1;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_244() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_245() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 11;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_246() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_247() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_248() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_249() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_250() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_251() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_252() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_253() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 9;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_254() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 6;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 6;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 10;
pd->face[2].orient[4] = 0;

pd->face[2].edge[5] = 11;
pd->face[2].orient[5] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_255() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_256() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 10;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_257() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_258() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_259() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_260() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_261() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_262() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_263() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_264() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 0;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 5;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_265() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_266() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_267() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_268() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_269() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_270() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_271() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_272() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_273() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_274() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_275() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_276() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_277() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_278() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_279() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_280() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 3;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 1;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_281() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_282() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 5;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_283() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_284() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_285() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_286() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 2;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_287() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_288() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_289() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_290() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_291() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_292() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_293() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 10;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_294() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_295() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_296() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_297() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_298() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_299() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_300() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_301() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_302() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_303() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_304() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 4;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_305() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_306() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_307() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_308() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_309() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_310() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 11;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_311() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_312() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 10;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_313() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 8;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_314() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_315() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_316() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_317() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_318() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_319() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_320() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_321() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_322() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_323() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_324() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 3;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_325() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_326() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_327() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 1;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_328() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_329() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_330() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_331() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_332() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_333() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_334() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_335() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_336() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_337() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_338() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_339() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_340() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[3].edge[3] = 6;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_341() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_342() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 1;
pd->face[3].orient[3] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_343() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_344() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_345() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 7;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_346() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 5;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_347() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_348() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_349() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_350() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_351() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 5;
pd->face[2].orient[4] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_352() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_353() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_354() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_355() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_356() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_357() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 7;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 6;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 1;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 1;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_358() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_359() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_360() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_361() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_362() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_363() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_364() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_365() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_366() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_367() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_368() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_369() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_370() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_371() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_372() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_373() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_374() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_375() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_376() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_377() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_378() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_379() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_380() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_381() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_382() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_383() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_384() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_385() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_386() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_387() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 2;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_388() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_389() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_390() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_391() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 8;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_392() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_393() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 9;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_394() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_395() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_396() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_397() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 6;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_398() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 6;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_399() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_400() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_401() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_402() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_403() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_404() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_405() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_406() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 7;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_407() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_408() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_409() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_410() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_411() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_412() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 7;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_413() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 4;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 2;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;

pd->face[7].edge[1] = 9;
pd->face[7].orient[1] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_414() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 8;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_415() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_416() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_417() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_418() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_419() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 2;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_420() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 0;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 5;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_421() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 1;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_422() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_423() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_424() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 4;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_425() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_426() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_427() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_428() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_429() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_430() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_431() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_432() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_433() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 0;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_434() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 3;
pd->cross[1].edge[3] = 7;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 1;
pd->edge[3].headpos = 2;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 4;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 4;
pd->comp[1].edge[1] = 5;
pd->comp[1].edge[2] = 6;
pd->comp[1].edge[3] = 7;

pd->comp[2].nedges = 4;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 8;
pd->comp[2].edge[1] = 9;
pd->comp[2].edge[2] = 10;
pd->comp[2].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_435() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_436() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 6;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_437() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_438() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 9;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 0;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 4;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_439() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 3;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_440() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_441() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_442() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 5;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 3;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_443() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_444() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_445() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_446() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_447() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_448() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 6;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_449() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_450() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_451() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_452() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 7;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_453() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 4;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 0;

pd->face[3].edge[3] = 5;
pd->face[3].orient[3] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_454() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 9;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_455() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_456() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_457() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_458() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_459() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_460() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_461() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 5;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_462() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_463() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_464() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_465() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_466() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_467() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_468() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_469() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_470() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 11;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_471() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_472() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_473() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 11;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 10;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 2;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 3;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_474() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_475() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_476() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_477() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_478() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_479() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_480() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_481() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_482() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_483() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_484() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_485() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_486() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_487() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_488() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 1;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 2;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_489() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 1;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 2;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_490() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_491() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 6;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_492() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 3;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_493() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_494() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 9;
pd->face[2].orient[3] = 1;

pd->face[2].edge[4] = 11;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_495() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_496() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 3;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 10;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 11;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 3;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 2;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 2;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 3;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;

pd->comp[2].nedges = 2;
pd->comp[2].tag = 'C';

pd->comp[2].edge = calloc(pd->comp[2].nedges,sizeof(pd_idx_t));
assert(pd->comp[2].edge != NULL);

pd->comp[2].edge[0] = 10;
pd->comp[2].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 9;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_497() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 2;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_498() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_499() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 10;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_500() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 8;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_501() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_502() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_503() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_504() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 7;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 0;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 0;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 0;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 2;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_505() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_506() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_507() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_508() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_509() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_510() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_511() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 7;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_512() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_513() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_514() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_515() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 3;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_516() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_517() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_518() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_519() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_520() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_521() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_522() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_523() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_524() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 11;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 6;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_525() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_526() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_527() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_528() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_529() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_530() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 9;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_531() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_532() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 1;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_533() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_534() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_535() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_536() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_537() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_538() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 4;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_539() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 3;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_540() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 3;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_541() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_542() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_543() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_544() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 1;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_545() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 8;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 1;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_546() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 11;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 10;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 2;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 2;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 7;
pd->face[6].orient[1] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_547() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_548() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 0;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 0;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_549() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_550() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_551() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 4;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_552() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_553() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[4].edge[2] = 9;
pd->face[4].orient[2] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_554() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 3;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 3;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[4].edge[2] = 7;
pd->face[4].orient[2] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_555() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 9;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_556() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_557() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_558() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_559() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_560() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_561() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_562() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 9;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 3;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 3;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_563() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 1;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_564() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_565() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_566() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 8;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_567() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 7;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_568() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_569() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_570() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_571() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_572() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_573() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_574() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_575() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_576() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_577() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_578() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 11;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 1;
pd->cross[1].edge[3] = 3;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 1;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 0;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 1;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 0;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 1;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 0;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 2;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 1;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 0;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 10;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_579() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 8;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 3;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_580() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 1;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_581() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_582() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 3;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_583() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_584() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 9;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_585() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_586() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_587() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_588() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_589() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_590() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 11;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_591() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_592() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_593() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_594() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_595() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_596() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_597() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_598() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 6;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_599() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 8;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_600() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_601() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 5;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[2].edge[4] = 4;
pd->face[2].orient[4] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_602() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 4;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 5;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 5;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 6;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 1;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 6;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 11;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 5;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 7;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_603() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_604() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_605() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_606() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_607() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 2;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_608() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_609() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_610() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_611() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 5;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 6;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 9;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_612() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_613() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_614() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_615() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_616() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_617() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_618() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_619() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_620() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 7;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_621() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 4;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 3;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 3;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_622() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_623() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_624() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 11;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_625() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 5;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 10;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_626() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_627() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_628() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_629() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_630() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 9;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 8;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 1;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_631() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 1;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_632() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 2;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 1;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 8;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 10;
pd->face[1].orient[7] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_633() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 9;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 3;
pd->face[1].orient[7] = 1;

pd->face[1].edge[8] = 2;
pd->face[1].orient[8] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_634() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 9;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 7;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 6;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 5;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 4;
pd->face[1].orient[7] = 0;

pd->face[1].edge[8] = 2;
pd->face[1].orient[8] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_635() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 9;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[1].nedges = 9;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[1].edge[8] = 1;
pd->face[1].orient[8] = 1;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_636() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 2;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[6].edge[1] = 11;
pd->face[6].orient[1] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 4;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_637() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_638() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 3;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 4;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 0;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 0;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 3;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 11;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 3;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 8;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 10;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_639() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 0;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 0;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 6;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 5;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_640() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 11;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 2;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_641() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_642() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 6;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 8;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 0;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_643() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 5;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_644() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 6;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 5;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 0;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 0;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 5;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 10;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 8;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[3].edge[2] = 10;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_645() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 9;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 11;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 1;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 5;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_646() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 4;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 2;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 8;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 4;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 8;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_647() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_648() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 1;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_649() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_650() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 5;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 4;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 11;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 10;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 1;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 1;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 10;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 5;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 11;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 8;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 10;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_651() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 7;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_652() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 4;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_653() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 8;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 9;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 10;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 11;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 3;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 3;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 11;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_654() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_655() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_656() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_657() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 9;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_658() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 11;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 11;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_659() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_660() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 2;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_661() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_662() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 3;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_663() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 6;
pd->cross[1].edge[2] = 5;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 7;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 8;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 1;
pd->edge[5].headpos = 2;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 1;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 4;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 3;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 6;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;

pd->comp[1].nedges = 6;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 6;
pd->comp[1].edge[1] = 7;
pd->comp[1].edge[2] = 8;
pd->comp[1].edge[3] = 9;
pd->comp[1].edge[4] = 10;
pd->comp[1].edge[5] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 11;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 5;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_664() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_665() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 9;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 11;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 1;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 3;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 6;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_666() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 11;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 0;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 0;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 1;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 6;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 7;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 2;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 8;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_667() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 8;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 2;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_668() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_669() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 9;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_670() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 9;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_671() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_672() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_673() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_674() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 7;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 11;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 1;
pd->edge[7].headpos = 2;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 1;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 4;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 4;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 8;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;

pd->comp[1].nedges = 4;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 8;
pd->comp[1].edge[1] = 9;
pd->comp[1].edge[2] = 10;
pd->comp[1].edge[3] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 8;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 10;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 7;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 9;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_675() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 3;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 5;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 2;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 10;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_676() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[2].edge[3] = 4;
pd->face[2].orient[3] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_677() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 2;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_678() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_679() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_680() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_681() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_682() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 11;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_683() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 2;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 9;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 1;
pd->edge[9].headpos = 2;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 1;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 1;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 10;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;

pd->comp[1].nedges = 2;
pd->comp[1].tag = 'B';

pd->comp[1].edge = calloc(pd->comp[1].nedges,sizeof(pd_idx_t));
assert(pd->comp[1].edge != NULL);

pd->comp[1].edge[0] = 10;
pd->comp[1].edge[1] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 9;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 7;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_684() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 11;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_685() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 9;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 10;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_686() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_687() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_688() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 4;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 7;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 6;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 8;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 3;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_689() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_690() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 1;
pd->face[1].orient[6] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_691() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 3;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_692() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 1;

pd->face[1].edge[5] = 8;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 10;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 11;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_693() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 5;
pd->face[1].orient[5] = 1;

pd->face[1].edge[6] = 4;
pd->face[1].orient[6] = 0;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_694() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 10;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 5;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 10;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[1].nedges = 8;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 3;
pd->face[1].orient[6] = 1;

pd->face[1].edge[7] = 2;
pd->face[1].orient[7] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_695() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 5;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 4;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 5;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[5].edge[1] = 7;
pd->face[5].orient[1] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_696() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 6;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 3;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 9;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_697() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 10;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 9;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 8;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 9;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 7;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 2;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 0;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 9;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 1;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 11;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_698() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 6;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 3;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 0;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 10;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_699() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 4;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 6;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_700() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 4;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 3;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 5;
pd->face[3].orient[1] = 0;

pd->face[3].edge[2] = 6;
pd->face[3].orient[2] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_701() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_702() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 3;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 4;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_703() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 4;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 11;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_704() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 1;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 2;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 4;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_705() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_706() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 7;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 4;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_707() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 2;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 1;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 3;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 3;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 5;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 4;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 7;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 7;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 3;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 8;
pd->face[2].orient[1] = 1;

pd->face[2].edge[2] = 10;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_708() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 1;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 6;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 8;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_709() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 1;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 4;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_710() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 1;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 4;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 6;
pd->face[2].orient[2] = 1;

pd->face[2].edge[3] = 5;
pd->face[2].orient[3] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_711() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_712() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 4;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 1;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_713() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 11;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 9;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 7;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_714() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 5;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 9;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 1;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_715() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 1;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 2;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 10;
pd->face[1].orient[4] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_716() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 10;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 9;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 8;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 3;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 4;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 6;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 7;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 2;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 1;
pd->face[0].orient[10] = 1;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[1].edge[4] = 11;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_717() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 7;
pd->face[2].orient[2] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_718() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_719() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 2;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 3;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 4;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 7;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 6;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 5;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 8;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 9;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 1;
pd->face[1].orient[5] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_720() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 11;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 10;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 11;
pd->face[0].orient[10] = 0;

pd->face[1].nedges = 7;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 9;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 8;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 6;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 4;
pd->face[1].orient[5] = 0;

pd->face[1].edge[6] = 2;
pd->face[1].orient[6] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_721() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 8;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 4;
pd->cross[5].edge[1] = 5;
pd->cross[5].edge[2] = 5;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 5;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 5;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 5;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 3;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 9;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 2;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 9;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 8;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 2;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[5].edge[1] = 6;
pd->face[5].orient[1] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 5;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_722() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 6;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 3;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 4;
pd->cross[4].edge[3] = 4;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 4;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 4;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 3;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 2;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 5;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 7;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 5;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_723() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 9;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 5;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 6;
pd->cross[5].edge[3] = 6;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 5;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 5;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 5;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 4;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 2;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 2;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 3;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 9;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 2;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 4;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 0;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 0;

pd->face[4].edge[1] = 7;
pd->face[4].orient[1] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 6;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_724() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 9;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 10;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 5;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 4;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 3;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 9;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 8;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 6;
pd->cross[5].edge[1] = 7;
pd->cross[5].edge[2] = 7;
pd->cross[5].edge[3] = 8;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 1;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 2;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 3;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 2;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 5;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 5;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 5;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 4;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 0;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 4;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 0;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 1;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 6;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 9;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 1;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 9;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 5;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 0;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 10;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 2;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 4;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 2;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 6;
pd->face[4].orient[0] = 1;

pd->face[4].edge[1] = 8;
pd->face[4].orient[1] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 3;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_725() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 7;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 8;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 3;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 2;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 3;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 1;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 0;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 1;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 2;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 3;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 5;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 4;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 7;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 10;
pd->face[1].orient[2] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 3;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 2;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_726() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 8;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 7;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 1;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 2;
pd->cross[2].edge[3] = 7;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 2;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 3;
pd->cross[3].edge[3] = 3;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 5;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 2;
pd->edge[1].headpos = 0;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 3;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 2;
pd->edge[2].tailpos = 2;

pd->edge[3].head = 3;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 3;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 2;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 0;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 2;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 0;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 7;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 1;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[0].edge[11] = 11;
pd->face[0].orient[11] = 0;

pd->face[1].nedges = 3;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 2;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 6;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 4;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 7;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_727() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 7;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 6;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 4;
pd->cross[4].edge[1] = 5;
pd->cross[4].edge[2] = 5;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 4;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 1;
pd->edge[5].tail = 4;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 3;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 3;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 3;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 4;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 1;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 4;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 6;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 0;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 5;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_728() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 10;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 9;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 7;
pd->cross[5].edge[1] = 8;
pd->cross[5].edge[2] = 8;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 5;
pd->edge[7].headpos = 0;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 1;
pd->edge[8].tail = 5;
pd->edge[8].tailpos = 2;

pd->edge[9].head = 4;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 3;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 4;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 9;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[0].edge[11] = 11;
pd->face[0].orient[11] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 7;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 9;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 8;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_729() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 9;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 8;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 6;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 5;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 4;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 9;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 10;
pd->cross[5].edge[3] = 11;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 1;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 2;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 3;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 2;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 1;
pd->edge[8].headpos = 3;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 0;
pd->edge[9].tail = 1;
pd->edge[9].tailpos = 1;

pd->edge[10].head = 5;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 2;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 5;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 5;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 11;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[0].edge[11] = 9;
pd->face[0].orient[11] = 0;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 8;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 6;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 2;
pd->face[1].orient[3] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 5;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 2;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 9;
pd->face[3].orient[0] = 1;

pd->face[3].edge[1] = 11;
pd->face[3].orient[1] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 1;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 4;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 7;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 10;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_730() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 1;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 2;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 10;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 11;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 7;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 8;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 3;
pd->cross[3].edge[1] = 5;
pd->cross[3].edge[2] = 4;
pd->cross[3].edge[3] = 4;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 7;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 6;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 9;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 10;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 1;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 3;

pd->edge[3].head = 3;
pd->edge[3].headpos = 0;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 3;
pd->edge[4].tail = 3;
pd->edge[4].tailpos = 2;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 1;

pd->edge[6].head = 4;
pd->edge[6].headpos = 3;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 2;
pd->edge[7].headpos = 1;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 1;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 2;
pd->edge[8].tailpos = 3;

pd->edge[9].head = 5;
pd->edge[9].headpos = 1;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 1;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 3;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 3;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 0;

pd->face[0].edge[1] = 11;
pd->face[0].orient[1] = 1;

pd->face[0].edge[2] = 10;
pd->face[0].orient[2] = 0;

pd->face[0].edge[3] = 9;
pd->face[0].orient[3] = 1;

pd->face[0].edge[4] = 8;
pd->face[0].orient[4] = 0;

pd->face[0].edge[5] = 3;
pd->face[0].orient[5] = 1;

pd->face[0].edge[6] = 4;
pd->face[0].orient[6] = 0;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 1;

pd->face[0].edge[8] = 6;
pd->face[0].orient[8] = 0;

pd->face[0].edge[9] = 7;
pd->face[0].orient[9] = 1;

pd->face[0].edge[10] = 2;
pd->face[0].orient[10] = 0;

pd->face[0].edge[11] = 1;
pd->face[0].orient[11] = 1;

pd->face[1].nedges = 4;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 1;

pd->face[1].edge[1] = 2;
pd->face[1].orient[1] = 1;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 1;

pd->face[1].edge[3] = 10;
pd->face[1].orient[3] = 1;

pd->face[2].nedges = 3;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 3;
pd->face[2].orient[0] = 0;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 0;

pd->face[2].edge[2] = 5;
pd->face[2].orient[2] = 0;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 0;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 4;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 0;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 0;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_731() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 8;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 7;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 5;
pd->cross[4].edge[1] = 6;
pd->cross[4].edge[2] = 6;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 4;
pd->edge[5].headpos = 0;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 1;
pd->edge[6].tail = 4;
pd->edge[6].tailpos = 2;

pd->edge[7].head = 3;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 3;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 3;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 7;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 5;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[0].edge[11] = 11;
pd->face[0].orient[11] = 0;

pd->face[1].nedges = 5;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 4;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 2;
pd->face[1].orient[4] = 0;

pd->face[2].nedges = 2;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 5;
pd->face[2].orient[0] = 1;

pd->face[2].edge[1] = 7;
pd->face[2].orient[1] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 1;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 3;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 6;
pd->face[5].orient[0] = 0;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

pd_code_t *pd_create_6_732() { 

/* This procedure is machine generated by pd_write_c */
/* and probably shouldn't be hand-edited. */

pd_code_t *pd;
pd = pd_code_new(6);
assert(pd != NULL);
pd->ncross = 6;
pd->nedges = 12;
pd->ncomps = 1;
pd->nfaces = 8;

/* Crossing data. */

pd->cross[0].edge[0] = 0;
pd->cross[0].edge[1] = 2;
pd->cross[0].edge[2] = 1;
pd->cross[0].edge[3] = 1;
pd->cross[0].sign = 2;

pd->cross[1].edge[0] = 0;
pd->cross[1].edge[1] = 11;
pd->cross[1].edge[2] = 11;
pd->cross[1].edge[3] = 10;
pd->cross[1].sign = 2;

pd->cross[2].edge[0] = 2;
pd->cross[2].edge[1] = 4;
pd->cross[2].edge[2] = 3;
pd->cross[2].edge[3] = 3;
pd->cross[2].sign = 2;

pd->cross[3].edge[0] = 4;
pd->cross[3].edge[1] = 6;
pd->cross[3].edge[2] = 5;
pd->cross[3].edge[3] = 5;
pd->cross[3].sign = 2;

pd->cross[4].edge[0] = 6;
pd->cross[4].edge[1] = 8;
pd->cross[4].edge[2] = 7;
pd->cross[4].edge[3] = 7;
pd->cross[4].sign = 2;

pd->cross[5].edge[0] = 8;
pd->cross[5].edge[1] = 10;
pd->cross[5].edge[2] = 9;
pd->cross[5].edge[3] = 9;
pd->cross[5].sign = 2;


/* Edge data */

pd->edge[0].head = 0;
pd->edge[0].headpos = 0;
pd->edge[0].tail = 1;
pd->edge[0].tailpos = 0;

pd->edge[1].head = 0;
pd->edge[1].headpos = 3;
pd->edge[1].tail = 0;
pd->edge[1].tailpos = 2;

pd->edge[2].head = 2;
pd->edge[2].headpos = 0;
pd->edge[2].tail = 0;
pd->edge[2].tailpos = 1;

pd->edge[3].head = 2;
pd->edge[3].headpos = 3;
pd->edge[3].tail = 2;
pd->edge[3].tailpos = 2;

pd->edge[4].head = 3;
pd->edge[4].headpos = 0;
pd->edge[4].tail = 2;
pd->edge[4].tailpos = 1;

pd->edge[5].head = 3;
pd->edge[5].headpos = 3;
pd->edge[5].tail = 3;
pd->edge[5].tailpos = 2;

pd->edge[6].head = 4;
pd->edge[6].headpos = 0;
pd->edge[6].tail = 3;
pd->edge[6].tailpos = 1;

pd->edge[7].head = 4;
pd->edge[7].headpos = 3;
pd->edge[7].tail = 4;
pd->edge[7].tailpos = 2;

pd->edge[8].head = 5;
pd->edge[8].headpos = 0;
pd->edge[8].tail = 4;
pd->edge[8].tailpos = 1;

pd->edge[9].head = 5;
pd->edge[9].headpos = 3;
pd->edge[9].tail = 5;
pd->edge[9].tailpos = 2;

pd->edge[10].head = 1;
pd->edge[10].headpos = 3;
pd->edge[10].tail = 5;
pd->edge[10].tailpos = 1;

pd->edge[11].head = 1;
pd->edge[11].headpos = 2;
pd->edge[11].tail = 1;
pd->edge[11].tailpos = 1;


/* Component Data */

pd->comp[0].nedges = 12;
pd->comp[0].tag = 'A';

pd->comp[0].edge = calloc(pd->comp[0].nedges,sizeof(pd_idx_t));
assert(pd->comp[0].edge != NULL);

pd->comp[0].edge[0] = 0;
pd->comp[0].edge[1] = 1;
pd->comp[0].edge[2] = 2;
pd->comp[0].edge[3] = 3;
pd->comp[0].edge[4] = 4;
pd->comp[0].edge[5] = 5;
pd->comp[0].edge[6] = 6;
pd->comp[0].edge[7] = 7;
pd->comp[0].edge[8] = 8;
pd->comp[0].edge[9] = 9;
pd->comp[0].edge[10] = 10;
pd->comp[0].edge[11] = 11;


/* Face data */

pd->face[0].nedges = 12;
pd->face[0].edge = calloc(pd->face[0].nedges,sizeof(pd_idx_t));
pd->face[0].orient = calloc(pd->face[0].nedges,sizeof(pd_or_t));
assert(pd->face[0].edge != NULL);
assert(pd->face[0].orient != NULL);

pd->face[0].edge[0] = 0;
pd->face[0].orient[0] = 1;

pd->face[0].edge[1] = 1;
pd->face[0].orient[1] = 0;

pd->face[0].edge[2] = 2;
pd->face[0].orient[2] = 1;

pd->face[0].edge[3] = 3;
pd->face[0].orient[3] = 0;

pd->face[0].edge[4] = 4;
pd->face[0].orient[4] = 1;

pd->face[0].edge[5] = 5;
pd->face[0].orient[5] = 0;

pd->face[0].edge[6] = 6;
pd->face[0].orient[6] = 1;

pd->face[0].edge[7] = 7;
pd->face[0].orient[7] = 0;

pd->face[0].edge[8] = 8;
pd->face[0].orient[8] = 1;

pd->face[0].edge[9] = 9;
pd->face[0].orient[9] = 0;

pd->face[0].edge[10] = 10;
pd->face[0].orient[10] = 1;

pd->face[0].edge[11] = 11;
pd->face[0].orient[11] = 0;

pd->face[1].nedges = 6;
pd->face[1].edge = calloc(pd->face[1].nedges,sizeof(pd_idx_t));
pd->face[1].orient = calloc(pd->face[1].nedges,sizeof(pd_or_t));
assert(pd->face[1].edge != NULL);
assert(pd->face[1].orient != NULL);

pd->face[1].edge[0] = 0;
pd->face[1].orient[0] = 0;

pd->face[1].edge[1] = 10;
pd->face[1].orient[1] = 0;

pd->face[1].edge[2] = 8;
pd->face[1].orient[2] = 0;

pd->face[1].edge[3] = 6;
pd->face[1].orient[3] = 0;

pd->face[1].edge[4] = 4;
pd->face[1].orient[4] = 0;

pd->face[1].edge[5] = 2;
pd->face[1].orient[5] = 0;

pd->face[2].nedges = 1;
pd->face[2].edge = calloc(pd->face[2].nedges,sizeof(pd_idx_t));
pd->face[2].orient = calloc(pd->face[2].nedges,sizeof(pd_or_t));
assert(pd->face[2].edge != NULL);
assert(pd->face[2].orient != NULL);

pd->face[2].edge[0] = 1;
pd->face[2].orient[0] = 1;

pd->face[3].nedges = 1;
pd->face[3].edge = calloc(pd->face[3].nedges,sizeof(pd_idx_t));
pd->face[3].orient = calloc(pd->face[3].nedges,sizeof(pd_or_t));
assert(pd->face[3].edge != NULL);
assert(pd->face[3].orient != NULL);

pd->face[3].edge[0] = 3;
pd->face[3].orient[0] = 1;

pd->face[4].nedges = 1;
pd->face[4].edge = calloc(pd->face[4].nedges,sizeof(pd_idx_t));
pd->face[4].orient = calloc(pd->face[4].nedges,sizeof(pd_or_t));
assert(pd->face[4].edge != NULL);
assert(pd->face[4].orient != NULL);

pd->face[4].edge[0] = 5;
pd->face[4].orient[0] = 1;

pd->face[5].nedges = 1;
pd->face[5].edge = calloc(pd->face[5].nedges,sizeof(pd_idx_t));
pd->face[5].orient = calloc(pd->face[5].nedges,sizeof(pd_or_t));
assert(pd->face[5].edge != NULL);
assert(pd->face[5].orient != NULL);

pd->face[5].edge[0] = 7;
pd->face[5].orient[0] = 1;

pd->face[6].nedges = 1;
pd->face[6].edge = calloc(pd->face[6].nedges,sizeof(pd_idx_t));
pd->face[6].orient = calloc(pd->face[6].nedges,sizeof(pd_or_t));
assert(pd->face[6].edge != NULL);
assert(pd->face[6].orient != NULL);

pd->face[6].edge[0] = 9;
pd->face[6].orient[0] = 1;

pd->face[7].nedges = 1;
pd->face[7].edge = calloc(pd->face[7].nedges,sizeof(pd_idx_t));
pd->face[7].orient = calloc(pd->face[7].nedges,sizeof(pd_or_t));
assert(pd->face[7].edge != NULL);
assert(pd->face[7].orient != NULL);

pd->face[7].edge[0] = 11;
pd->face[7].orient[0] = 1;


/* End of data. */

pd_regenerate_hash(pd);
assert(pd_ok(pd));
return pd;

}

bool test_diagrams_6() {
pd_code_t *pd;
printf("testing diagram pd_create_6_0...");
pd = pd_create_6_0();
if (!all_diagram_isotopies_test(pd,"pd_create_6_0")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_1...");
pd = pd_create_6_1();
if (!all_diagram_isotopies_test(pd,"pd_create_6_1")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_10...");
pd = pd_create_6_10();
if (!all_diagram_isotopies_test(pd,"pd_create_6_10")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_100...");
pd = pd_create_6_100();
if (!all_diagram_isotopies_test(pd,"pd_create_6_100")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_101...");
pd = pd_create_6_101();
if (!all_diagram_isotopies_test(pd,"pd_create_6_101")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_102...");
pd = pd_create_6_102();
if (!all_diagram_isotopies_test(pd,"pd_create_6_102")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_103...");
pd = pd_create_6_103();
if (!all_diagram_isotopies_test(pd,"pd_create_6_103")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_104...");
pd = pd_create_6_104();
if (!all_diagram_isotopies_test(pd,"pd_create_6_104")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_105...");
pd = pd_create_6_105();
if (!all_diagram_isotopies_test(pd,"pd_create_6_105")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_106...");
pd = pd_create_6_106();
if (!all_diagram_isotopies_test(pd,"pd_create_6_106")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_107...");
pd = pd_create_6_107();
if (!all_diagram_isotopies_test(pd,"pd_create_6_107")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_108...");
pd = pd_create_6_108();
if (!all_diagram_isotopies_test(pd,"pd_create_6_108")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_109...");
pd = pd_create_6_109();
if (!all_diagram_isotopies_test(pd,"pd_create_6_109")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_11...");
pd = pd_create_6_11();
if (!all_diagram_isotopies_test(pd,"pd_create_6_11")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_110...");
pd = pd_create_6_110();
if (!all_diagram_isotopies_test(pd,"pd_create_6_110")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_111...");
pd = pd_create_6_111();
if (!all_diagram_isotopies_test(pd,"pd_create_6_111")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_112...");
pd = pd_create_6_112();
if (!all_diagram_isotopies_test(pd,"pd_create_6_112")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_113...");
pd = pd_create_6_113();
if (!all_diagram_isotopies_test(pd,"pd_create_6_113")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_114...");
pd = pd_create_6_114();
if (!all_diagram_isotopies_test(pd,"pd_create_6_114")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_115...");
pd = pd_create_6_115();
if (!all_diagram_isotopies_test(pd,"pd_create_6_115")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_116...");
pd = pd_create_6_116();
if (!all_diagram_isotopies_test(pd,"pd_create_6_116")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_117...");
pd = pd_create_6_117();
if (!all_diagram_isotopies_test(pd,"pd_create_6_117")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_118...");
pd = pd_create_6_118();
if (!all_diagram_isotopies_test(pd,"pd_create_6_118")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_119...");
pd = pd_create_6_119();
if (!all_diagram_isotopies_test(pd,"pd_create_6_119")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_12...");
pd = pd_create_6_12();
if (!all_diagram_isotopies_test(pd,"pd_create_6_12")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_120...");
pd = pd_create_6_120();
if (!all_diagram_isotopies_test(pd,"pd_create_6_120")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_121...");
pd = pd_create_6_121();
if (!all_diagram_isotopies_test(pd,"pd_create_6_121")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_122...");
pd = pd_create_6_122();
if (!all_diagram_isotopies_test(pd,"pd_create_6_122")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_123...");
pd = pd_create_6_123();
if (!all_diagram_isotopies_test(pd,"pd_create_6_123")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_124...");
pd = pd_create_6_124();
if (!all_diagram_isotopies_test(pd,"pd_create_6_124")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_125...");
pd = pd_create_6_125();
if (!all_diagram_isotopies_test(pd,"pd_create_6_125")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_126...");
pd = pd_create_6_126();
if (!all_diagram_isotopies_test(pd,"pd_create_6_126")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_127...");
pd = pd_create_6_127();
if (!all_diagram_isotopies_test(pd,"pd_create_6_127")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_128...");
pd = pd_create_6_128();
if (!all_diagram_isotopies_test(pd,"pd_create_6_128")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_129...");
pd = pd_create_6_129();
if (!all_diagram_isotopies_test(pd,"pd_create_6_129")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_13...");
pd = pd_create_6_13();
if (!all_diagram_isotopies_test(pd,"pd_create_6_13")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_130...");
pd = pd_create_6_130();
if (!all_diagram_isotopies_test(pd,"pd_create_6_130")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_131...");
pd = pd_create_6_131();
if (!all_diagram_isotopies_test(pd,"pd_create_6_131")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_132...");
pd = pd_create_6_132();
if (!all_diagram_isotopies_test(pd,"pd_create_6_132")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_133...");
pd = pd_create_6_133();
if (!all_diagram_isotopies_test(pd,"pd_create_6_133")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_134...");
pd = pd_create_6_134();
if (!all_diagram_isotopies_test(pd,"pd_create_6_134")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_135...");
pd = pd_create_6_135();
if (!all_diagram_isotopies_test(pd,"pd_create_6_135")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_136...");
pd = pd_create_6_136();
if (!all_diagram_isotopies_test(pd,"pd_create_6_136")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_137...");
pd = pd_create_6_137();
if (!all_diagram_isotopies_test(pd,"pd_create_6_137")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_138...");
pd = pd_create_6_138();
if (!all_diagram_isotopies_test(pd,"pd_create_6_138")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_139...");
pd = pd_create_6_139();
if (!all_diagram_isotopies_test(pd,"pd_create_6_139")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_14...");
pd = pd_create_6_14();
if (!all_diagram_isotopies_test(pd,"pd_create_6_14")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_140...");
pd = pd_create_6_140();
if (!all_diagram_isotopies_test(pd,"pd_create_6_140")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_141...");
pd = pd_create_6_141();
if (!all_diagram_isotopies_test(pd,"pd_create_6_141")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_142...");
pd = pd_create_6_142();
if (!all_diagram_isotopies_test(pd,"pd_create_6_142")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_143...");
pd = pd_create_6_143();
if (!all_diagram_isotopies_test(pd,"pd_create_6_143")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_144...");
pd = pd_create_6_144();
if (!all_diagram_isotopies_test(pd,"pd_create_6_144")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_145...");
pd = pd_create_6_145();
if (!all_diagram_isotopies_test(pd,"pd_create_6_145")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_146...");
pd = pd_create_6_146();
if (!all_diagram_isotopies_test(pd,"pd_create_6_146")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_147...");
pd = pd_create_6_147();
if (!all_diagram_isotopies_test(pd,"pd_create_6_147")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_148...");
pd = pd_create_6_148();
if (!all_diagram_isotopies_test(pd,"pd_create_6_148")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_149...");
pd = pd_create_6_149();
if (!all_diagram_isotopies_test(pd,"pd_create_6_149")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_15...");
pd = pd_create_6_15();
if (!all_diagram_isotopies_test(pd,"pd_create_6_15")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_150...");
pd = pd_create_6_150();
if (!all_diagram_isotopies_test(pd,"pd_create_6_150")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_151...");
pd = pd_create_6_151();
if (!all_diagram_isotopies_test(pd,"pd_create_6_151")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_152...");
pd = pd_create_6_152();
if (!all_diagram_isotopies_test(pd,"pd_create_6_152")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_153...");
pd = pd_create_6_153();
if (!all_diagram_isotopies_test(pd,"pd_create_6_153")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_154...");
pd = pd_create_6_154();
if (!all_diagram_isotopies_test(pd,"pd_create_6_154")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_155...");
pd = pd_create_6_155();
if (!all_diagram_isotopies_test(pd,"pd_create_6_155")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_156...");
pd = pd_create_6_156();
if (!all_diagram_isotopies_test(pd,"pd_create_6_156")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_157...");
pd = pd_create_6_157();
if (!all_diagram_isotopies_test(pd,"pd_create_6_157")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_158...");
pd = pd_create_6_158();
if (!all_diagram_isotopies_test(pd,"pd_create_6_158")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_159...");
pd = pd_create_6_159();
if (!all_diagram_isotopies_test(pd,"pd_create_6_159")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_16...");
pd = pd_create_6_16();
if (!all_diagram_isotopies_test(pd,"pd_create_6_16")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_160...");
pd = pd_create_6_160();
if (!all_diagram_isotopies_test(pd,"pd_create_6_160")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_161...");
pd = pd_create_6_161();
if (!all_diagram_isotopies_test(pd,"pd_create_6_161")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_162...");
pd = pd_create_6_162();
if (!all_diagram_isotopies_test(pd,"pd_create_6_162")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_163...");
pd = pd_create_6_163();
if (!all_diagram_isotopies_test(pd,"pd_create_6_163")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_164...");
pd = pd_create_6_164();
if (!all_diagram_isotopies_test(pd,"pd_create_6_164")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_165...");
pd = pd_create_6_165();
if (!all_diagram_isotopies_test(pd,"pd_create_6_165")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_166...");
pd = pd_create_6_166();
if (!all_diagram_isotopies_test(pd,"pd_create_6_166")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_167...");
pd = pd_create_6_167();
if (!all_diagram_isotopies_test(pd,"pd_create_6_167")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_168...");
pd = pd_create_6_168();
if (!all_diagram_isotopies_test(pd,"pd_create_6_168")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_169...");
pd = pd_create_6_169();
if (!all_diagram_isotopies_test(pd,"pd_create_6_169")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_17...");
pd = pd_create_6_17();
if (!all_diagram_isotopies_test(pd,"pd_create_6_17")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_170...");
pd = pd_create_6_170();
if (!all_diagram_isotopies_test(pd,"pd_create_6_170")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_171...");
pd = pd_create_6_171();
if (!all_diagram_isotopies_test(pd,"pd_create_6_171")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_172...");
pd = pd_create_6_172();
if (!all_diagram_isotopies_test(pd,"pd_create_6_172")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_173...");
pd = pd_create_6_173();
if (!all_diagram_isotopies_test(pd,"pd_create_6_173")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_174...");
pd = pd_create_6_174();
if (!all_diagram_isotopies_test(pd,"pd_create_6_174")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_175...");
pd = pd_create_6_175();
if (!all_diagram_isotopies_test(pd,"pd_create_6_175")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_176...");
pd = pd_create_6_176();
if (!all_diagram_isotopies_test(pd,"pd_create_6_176")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_177...");
pd = pd_create_6_177();
if (!all_diagram_isotopies_test(pd,"pd_create_6_177")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_178...");
pd = pd_create_6_178();
if (!all_diagram_isotopies_test(pd,"pd_create_6_178")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_179...");
pd = pd_create_6_179();
if (!all_diagram_isotopies_test(pd,"pd_create_6_179")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_18...");
pd = pd_create_6_18();
if (!all_diagram_isotopies_test(pd,"pd_create_6_18")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_180...");
pd = pd_create_6_180();
if (!all_diagram_isotopies_test(pd,"pd_create_6_180")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_181...");
pd = pd_create_6_181();
if (!all_diagram_isotopies_test(pd,"pd_create_6_181")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_182...");
pd = pd_create_6_182();
if (!all_diagram_isotopies_test(pd,"pd_create_6_182")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_183...");
pd = pd_create_6_183();
if (!all_diagram_isotopies_test(pd,"pd_create_6_183")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_184...");
pd = pd_create_6_184();
if (!all_diagram_isotopies_test(pd,"pd_create_6_184")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_185...");
pd = pd_create_6_185();
if (!all_diagram_isotopies_test(pd,"pd_create_6_185")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_186...");
pd = pd_create_6_186();
if (!all_diagram_isotopies_test(pd,"pd_create_6_186")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_187...");
pd = pd_create_6_187();
if (!all_diagram_isotopies_test(pd,"pd_create_6_187")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_188...");
pd = pd_create_6_188();
if (!all_diagram_isotopies_test(pd,"pd_create_6_188")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_189...");
pd = pd_create_6_189();
if (!all_diagram_isotopies_test(pd,"pd_create_6_189")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_19...");
pd = pd_create_6_19();
if (!all_diagram_isotopies_test(pd,"pd_create_6_19")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_190...");
pd = pd_create_6_190();
if (!all_diagram_isotopies_test(pd,"pd_create_6_190")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_191...");
pd = pd_create_6_191();
if (!all_diagram_isotopies_test(pd,"pd_create_6_191")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_192...");
pd = pd_create_6_192();
if (!all_diagram_isotopies_test(pd,"pd_create_6_192")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_193...");
pd = pd_create_6_193();
if (!all_diagram_isotopies_test(pd,"pd_create_6_193")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_194...");
pd = pd_create_6_194();
if (!all_diagram_isotopies_test(pd,"pd_create_6_194")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_195...");
pd = pd_create_6_195();
if (!all_diagram_isotopies_test(pd,"pd_create_6_195")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_196...");
pd = pd_create_6_196();
if (!all_diagram_isotopies_test(pd,"pd_create_6_196")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_197...");
pd = pd_create_6_197();
if (!all_diagram_isotopies_test(pd,"pd_create_6_197")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_198...");
pd = pd_create_6_198();
if (!all_diagram_isotopies_test(pd,"pd_create_6_198")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_199...");
pd = pd_create_6_199();
if (!all_diagram_isotopies_test(pd,"pd_create_6_199")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_2...");
pd = pd_create_6_2();
if (!all_diagram_isotopies_test(pd,"pd_create_6_2")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_20...");
pd = pd_create_6_20();
if (!all_diagram_isotopies_test(pd,"pd_create_6_20")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_200...");
pd = pd_create_6_200();
if (!all_diagram_isotopies_test(pd,"pd_create_6_200")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_201...");
pd = pd_create_6_201();
if (!all_diagram_isotopies_test(pd,"pd_create_6_201")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_202...");
pd = pd_create_6_202();
if (!all_diagram_isotopies_test(pd,"pd_create_6_202")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_203...");
pd = pd_create_6_203();
if (!all_diagram_isotopies_test(pd,"pd_create_6_203")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_204...");
pd = pd_create_6_204();
if (!all_diagram_isotopies_test(pd,"pd_create_6_204")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_205...");
pd = pd_create_6_205();
if (!all_diagram_isotopies_test(pd,"pd_create_6_205")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_206...");
pd = pd_create_6_206();
if (!all_diagram_isotopies_test(pd,"pd_create_6_206")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_207...");
pd = pd_create_6_207();
if (!all_diagram_isotopies_test(pd,"pd_create_6_207")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_208...");
pd = pd_create_6_208();
if (!all_diagram_isotopies_test(pd,"pd_create_6_208")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_209...");
pd = pd_create_6_209();
if (!all_diagram_isotopies_test(pd,"pd_create_6_209")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_21...");
pd = pd_create_6_21();
if (!all_diagram_isotopies_test(pd,"pd_create_6_21")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_210...");
pd = pd_create_6_210();
if (!all_diagram_isotopies_test(pd,"pd_create_6_210")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_211...");
pd = pd_create_6_211();
if (!all_diagram_isotopies_test(pd,"pd_create_6_211")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_212...");
pd = pd_create_6_212();
if (!all_diagram_isotopies_test(pd,"pd_create_6_212")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_213...");
pd = pd_create_6_213();
if (!all_diagram_isotopies_test(pd,"pd_create_6_213")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_214...");
pd = pd_create_6_214();
if (!all_diagram_isotopies_test(pd,"pd_create_6_214")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_215...");
pd = pd_create_6_215();
if (!all_diagram_isotopies_test(pd,"pd_create_6_215")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_216...");
pd = pd_create_6_216();
if (!all_diagram_isotopies_test(pd,"pd_create_6_216")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_217...");
pd = pd_create_6_217();
if (!all_diagram_isotopies_test(pd,"pd_create_6_217")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_218...");
pd = pd_create_6_218();
if (!all_diagram_isotopies_test(pd,"pd_create_6_218")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_219...");
pd = pd_create_6_219();
if (!all_diagram_isotopies_test(pd,"pd_create_6_219")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_22...");
pd = pd_create_6_22();
if (!all_diagram_isotopies_test(pd,"pd_create_6_22")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_220...");
pd = pd_create_6_220();
if (!all_diagram_isotopies_test(pd,"pd_create_6_220")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_221...");
pd = pd_create_6_221();
if (!all_diagram_isotopies_test(pd,"pd_create_6_221")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_222...");
pd = pd_create_6_222();
if (!all_diagram_isotopies_test(pd,"pd_create_6_222")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_223...");
pd = pd_create_6_223();
if (!all_diagram_isotopies_test(pd,"pd_create_6_223")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_224...");
pd = pd_create_6_224();
if (!all_diagram_isotopies_test(pd,"pd_create_6_224")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_225...");
pd = pd_create_6_225();
if (!all_diagram_isotopies_test(pd,"pd_create_6_225")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_226...");
pd = pd_create_6_226();
if (!all_diagram_isotopies_test(pd,"pd_create_6_226")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_227...");
pd = pd_create_6_227();
if (!all_diagram_isotopies_test(pd,"pd_create_6_227")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_228...");
pd = pd_create_6_228();
if (!all_diagram_isotopies_test(pd,"pd_create_6_228")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_229...");
pd = pd_create_6_229();
if (!all_diagram_isotopies_test(pd,"pd_create_6_229")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_23...");
pd = pd_create_6_23();
if (!all_diagram_isotopies_test(pd,"pd_create_6_23")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_230...");
pd = pd_create_6_230();
if (!all_diagram_isotopies_test(pd,"pd_create_6_230")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_231...");
pd = pd_create_6_231();
if (!all_diagram_isotopies_test(pd,"pd_create_6_231")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_232...");
pd = pd_create_6_232();
if (!all_diagram_isotopies_test(pd,"pd_create_6_232")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_233...");
pd = pd_create_6_233();
if (!all_diagram_isotopies_test(pd,"pd_create_6_233")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_234...");
pd = pd_create_6_234();
if (!all_diagram_isotopies_test(pd,"pd_create_6_234")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_235...");
pd = pd_create_6_235();
if (!all_diagram_isotopies_test(pd,"pd_create_6_235")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_236...");
pd = pd_create_6_236();
if (!all_diagram_isotopies_test(pd,"pd_create_6_236")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_237...");
pd = pd_create_6_237();
if (!all_diagram_isotopies_test(pd,"pd_create_6_237")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_238...");
pd = pd_create_6_238();
if (!all_diagram_isotopies_test(pd,"pd_create_6_238")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_239...");
pd = pd_create_6_239();
if (!all_diagram_isotopies_test(pd,"pd_create_6_239")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_24...");
pd = pd_create_6_24();
if (!all_diagram_isotopies_test(pd,"pd_create_6_24")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_240...");
pd = pd_create_6_240();
if (!all_diagram_isotopies_test(pd,"pd_create_6_240")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_241...");
pd = pd_create_6_241();
if (!all_diagram_isotopies_test(pd,"pd_create_6_241")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_242...");
pd = pd_create_6_242();
if (!all_diagram_isotopies_test(pd,"pd_create_6_242")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_243...");
pd = pd_create_6_243();
if (!all_diagram_isotopies_test(pd,"pd_create_6_243")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_244...");
pd = pd_create_6_244();
if (!all_diagram_isotopies_test(pd,"pd_create_6_244")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_245...");
pd = pd_create_6_245();
if (!all_diagram_isotopies_test(pd,"pd_create_6_245")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_246...");
pd = pd_create_6_246();
if (!all_diagram_isotopies_test(pd,"pd_create_6_246")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_247...");
pd = pd_create_6_247();
if (!all_diagram_isotopies_test(pd,"pd_create_6_247")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_248...");
pd = pd_create_6_248();
if (!all_diagram_isotopies_test(pd,"pd_create_6_248")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_249...");
pd = pd_create_6_249();
if (!all_diagram_isotopies_test(pd,"pd_create_6_249")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_25...");
pd = pd_create_6_25();
if (!all_diagram_isotopies_test(pd,"pd_create_6_25")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_250...");
pd = pd_create_6_250();
if (!all_diagram_isotopies_test(pd,"pd_create_6_250")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_251...");
pd = pd_create_6_251();
if (!all_diagram_isotopies_test(pd,"pd_create_6_251")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_252...");
pd = pd_create_6_252();
if (!all_diagram_isotopies_test(pd,"pd_create_6_252")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_253...");
pd = pd_create_6_253();
if (!all_diagram_isotopies_test(pd,"pd_create_6_253")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_254...");
pd = pd_create_6_254();
if (!all_diagram_isotopies_test(pd,"pd_create_6_254")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_255...");
pd = pd_create_6_255();
if (!all_diagram_isotopies_test(pd,"pd_create_6_255")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_256...");
pd = pd_create_6_256();
if (!all_diagram_isotopies_test(pd,"pd_create_6_256")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_257...");
pd = pd_create_6_257();
if (!all_diagram_isotopies_test(pd,"pd_create_6_257")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_258...");
pd = pd_create_6_258();
if (!all_diagram_isotopies_test(pd,"pd_create_6_258")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_259...");
pd = pd_create_6_259();
if (!all_diagram_isotopies_test(pd,"pd_create_6_259")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_26...");
pd = pd_create_6_26();
if (!all_diagram_isotopies_test(pd,"pd_create_6_26")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_260...");
pd = pd_create_6_260();
if (!all_diagram_isotopies_test(pd,"pd_create_6_260")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_261...");
pd = pd_create_6_261();
if (!all_diagram_isotopies_test(pd,"pd_create_6_261")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_262...");
pd = pd_create_6_262();
if (!all_diagram_isotopies_test(pd,"pd_create_6_262")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_263...");
pd = pd_create_6_263();
if (!all_diagram_isotopies_test(pd,"pd_create_6_263")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_264...");
pd = pd_create_6_264();
if (!all_diagram_isotopies_test(pd,"pd_create_6_264")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_265...");
pd = pd_create_6_265();
if (!all_diagram_isotopies_test(pd,"pd_create_6_265")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_266...");
pd = pd_create_6_266();
if (!all_diagram_isotopies_test(pd,"pd_create_6_266")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_267...");
pd = pd_create_6_267();
if (!all_diagram_isotopies_test(pd,"pd_create_6_267")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_268...");
pd = pd_create_6_268();
if (!all_diagram_isotopies_test(pd,"pd_create_6_268")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_269...");
pd = pd_create_6_269();
if (!all_diagram_isotopies_test(pd,"pd_create_6_269")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_27...");
pd = pd_create_6_27();
if (!all_diagram_isotopies_test(pd,"pd_create_6_27")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_270...");
pd = pd_create_6_270();
if (!all_diagram_isotopies_test(pd,"pd_create_6_270")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_271...");
pd = pd_create_6_271();
if (!all_diagram_isotopies_test(pd,"pd_create_6_271")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_272...");
pd = pd_create_6_272();
if (!all_diagram_isotopies_test(pd,"pd_create_6_272")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_273...");
pd = pd_create_6_273();
if (!all_diagram_isotopies_test(pd,"pd_create_6_273")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_274...");
pd = pd_create_6_274();
if (!all_diagram_isotopies_test(pd,"pd_create_6_274")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_275...");
pd = pd_create_6_275();
if (!all_diagram_isotopies_test(pd,"pd_create_6_275")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_276...");
pd = pd_create_6_276();
if (!all_diagram_isotopies_test(pd,"pd_create_6_276")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_277...");
pd = pd_create_6_277();
if (!all_diagram_isotopies_test(pd,"pd_create_6_277")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_278...");
pd = pd_create_6_278();
if (!all_diagram_isotopies_test(pd,"pd_create_6_278")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_279...");
pd = pd_create_6_279();
if (!all_diagram_isotopies_test(pd,"pd_create_6_279")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_28...");
pd = pd_create_6_28();
if (!all_diagram_isotopies_test(pd,"pd_create_6_28")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_280...");
pd = pd_create_6_280();
if (!all_diagram_isotopies_test(pd,"pd_create_6_280")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_281...");
pd = pd_create_6_281();
if (!all_diagram_isotopies_test(pd,"pd_create_6_281")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_282...");
pd = pd_create_6_282();
if (!all_diagram_isotopies_test(pd,"pd_create_6_282")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_283...");
pd = pd_create_6_283();
if (!all_diagram_isotopies_test(pd,"pd_create_6_283")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_284...");
pd = pd_create_6_284();
if (!all_diagram_isotopies_test(pd,"pd_create_6_284")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_285...");
pd = pd_create_6_285();
if (!all_diagram_isotopies_test(pd,"pd_create_6_285")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_286...");
pd = pd_create_6_286();
if (!all_diagram_isotopies_test(pd,"pd_create_6_286")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_287...");
pd = pd_create_6_287();
if (!all_diagram_isotopies_test(pd,"pd_create_6_287")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_288...");
pd = pd_create_6_288();
if (!all_diagram_isotopies_test(pd,"pd_create_6_288")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_289...");
pd = pd_create_6_289();
if (!all_diagram_isotopies_test(pd,"pd_create_6_289")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_29...");
pd = pd_create_6_29();
if (!all_diagram_isotopies_test(pd,"pd_create_6_29")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_290...");
pd = pd_create_6_290();
if (!all_diagram_isotopies_test(pd,"pd_create_6_290")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_291...");
pd = pd_create_6_291();
if (!all_diagram_isotopies_test(pd,"pd_create_6_291")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_292...");
pd = pd_create_6_292();
if (!all_diagram_isotopies_test(pd,"pd_create_6_292")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_293...");
pd = pd_create_6_293();
if (!all_diagram_isotopies_test(pd,"pd_create_6_293")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_294...");
pd = pd_create_6_294();
if (!all_diagram_isotopies_test(pd,"pd_create_6_294")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_295...");
pd = pd_create_6_295();
if (!all_diagram_isotopies_test(pd,"pd_create_6_295")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_296...");
pd = pd_create_6_296();
if (!all_diagram_isotopies_test(pd,"pd_create_6_296")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_297...");
pd = pd_create_6_297();
if (!all_diagram_isotopies_test(pd,"pd_create_6_297")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_298...");
pd = pd_create_6_298();
if (!all_diagram_isotopies_test(pd,"pd_create_6_298")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_299...");
pd = pd_create_6_299();
if (!all_diagram_isotopies_test(pd,"pd_create_6_299")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_3...");
pd = pd_create_6_3();
if (!all_diagram_isotopies_test(pd,"pd_create_6_3")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_30...");
pd = pd_create_6_30();
if (!all_diagram_isotopies_test(pd,"pd_create_6_30")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_300...");
pd = pd_create_6_300();
if (!all_diagram_isotopies_test(pd,"pd_create_6_300")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_301...");
pd = pd_create_6_301();
if (!all_diagram_isotopies_test(pd,"pd_create_6_301")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_302...");
pd = pd_create_6_302();
if (!all_diagram_isotopies_test(pd,"pd_create_6_302")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_303...");
pd = pd_create_6_303();
if (!all_diagram_isotopies_test(pd,"pd_create_6_303")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_304...");
pd = pd_create_6_304();
if (!all_diagram_isotopies_test(pd,"pd_create_6_304")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_305...");
pd = pd_create_6_305();
if (!all_diagram_isotopies_test(pd,"pd_create_6_305")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_306...");
pd = pd_create_6_306();
if (!all_diagram_isotopies_test(pd,"pd_create_6_306")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_307...");
pd = pd_create_6_307();
if (!all_diagram_isotopies_test(pd,"pd_create_6_307")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_308...");
pd = pd_create_6_308();
if (!all_diagram_isotopies_test(pd,"pd_create_6_308")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_309...");
pd = pd_create_6_309();
if (!all_diagram_isotopies_test(pd,"pd_create_6_309")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_31...");
pd = pd_create_6_31();
if (!all_diagram_isotopies_test(pd,"pd_create_6_31")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_310...");
pd = pd_create_6_310();
if (!all_diagram_isotopies_test(pd,"pd_create_6_310")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_311...");
pd = pd_create_6_311();
if (!all_diagram_isotopies_test(pd,"pd_create_6_311")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_312...");
pd = pd_create_6_312();
if (!all_diagram_isotopies_test(pd,"pd_create_6_312")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_313...");
pd = pd_create_6_313();
if (!all_diagram_isotopies_test(pd,"pd_create_6_313")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_314...");
pd = pd_create_6_314();
if (!all_diagram_isotopies_test(pd,"pd_create_6_314")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_315...");
pd = pd_create_6_315();
if (!all_diagram_isotopies_test(pd,"pd_create_6_315")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_316...");
pd = pd_create_6_316();
if (!all_diagram_isotopies_test(pd,"pd_create_6_316")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_317...");
pd = pd_create_6_317();
if (!all_diagram_isotopies_test(pd,"pd_create_6_317")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_318...");
pd = pd_create_6_318();
if (!all_diagram_isotopies_test(pd,"pd_create_6_318")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_319...");
pd = pd_create_6_319();
if (!all_diagram_isotopies_test(pd,"pd_create_6_319")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_32...");
pd = pd_create_6_32();
if (!all_diagram_isotopies_test(pd,"pd_create_6_32")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_320...");
pd = pd_create_6_320();
if (!all_diagram_isotopies_test(pd,"pd_create_6_320")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_321...");
pd = pd_create_6_321();
if (!all_diagram_isotopies_test(pd,"pd_create_6_321")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_322...");
pd = pd_create_6_322();
if (!all_diagram_isotopies_test(pd,"pd_create_6_322")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_323...");
pd = pd_create_6_323();
if (!all_diagram_isotopies_test(pd,"pd_create_6_323")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_324...");
pd = pd_create_6_324();
if (!all_diagram_isotopies_test(pd,"pd_create_6_324")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_325...");
pd = pd_create_6_325();
if (!all_diagram_isotopies_test(pd,"pd_create_6_325")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_326...");
pd = pd_create_6_326();
if (!all_diagram_isotopies_test(pd,"pd_create_6_326")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_327...");
pd = pd_create_6_327();
if (!all_diagram_isotopies_test(pd,"pd_create_6_327")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_328...");
pd = pd_create_6_328();
if (!all_diagram_isotopies_test(pd,"pd_create_6_328")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_329...");
pd = pd_create_6_329();
if (!all_diagram_isotopies_test(pd,"pd_create_6_329")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_33...");
pd = pd_create_6_33();
if (!all_diagram_isotopies_test(pd,"pd_create_6_33")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_330...");
pd = pd_create_6_330();
if (!all_diagram_isotopies_test(pd,"pd_create_6_330")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_331...");
pd = pd_create_6_331();
if (!all_diagram_isotopies_test(pd,"pd_create_6_331")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_332...");
pd = pd_create_6_332();
if (!all_diagram_isotopies_test(pd,"pd_create_6_332")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_333...");
pd = pd_create_6_333();
if (!all_diagram_isotopies_test(pd,"pd_create_6_333")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_334...");
pd = pd_create_6_334();
if (!all_diagram_isotopies_test(pd,"pd_create_6_334")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_335...");
pd = pd_create_6_335();
if (!all_diagram_isotopies_test(pd,"pd_create_6_335")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_336...");
pd = pd_create_6_336();
if (!all_diagram_isotopies_test(pd,"pd_create_6_336")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_337...");
pd = pd_create_6_337();
if (!all_diagram_isotopies_test(pd,"pd_create_6_337")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_338...");
pd = pd_create_6_338();
if (!all_diagram_isotopies_test(pd,"pd_create_6_338")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_339...");
pd = pd_create_6_339();
if (!all_diagram_isotopies_test(pd,"pd_create_6_339")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_34...");
pd = pd_create_6_34();
if (!all_diagram_isotopies_test(pd,"pd_create_6_34")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_340...");
pd = pd_create_6_340();
if (!all_diagram_isotopies_test(pd,"pd_create_6_340")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_341...");
pd = pd_create_6_341();
if (!all_diagram_isotopies_test(pd,"pd_create_6_341")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_342...");
pd = pd_create_6_342();
if (!all_diagram_isotopies_test(pd,"pd_create_6_342")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_343...");
pd = pd_create_6_343();
if (!all_diagram_isotopies_test(pd,"pd_create_6_343")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_344...");
pd = pd_create_6_344();
if (!all_diagram_isotopies_test(pd,"pd_create_6_344")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_345...");
pd = pd_create_6_345();
if (!all_diagram_isotopies_test(pd,"pd_create_6_345")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_346...");
pd = pd_create_6_346();
if (!all_diagram_isotopies_test(pd,"pd_create_6_346")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_347...");
pd = pd_create_6_347();
if (!all_diagram_isotopies_test(pd,"pd_create_6_347")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_348...");
pd = pd_create_6_348();
if (!all_diagram_isotopies_test(pd,"pd_create_6_348")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_349...");
pd = pd_create_6_349();
if (!all_diagram_isotopies_test(pd,"pd_create_6_349")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_35...");
pd = pd_create_6_35();
if (!all_diagram_isotopies_test(pd,"pd_create_6_35")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_350...");
pd = pd_create_6_350();
if (!all_diagram_isotopies_test(pd,"pd_create_6_350")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_351...");
pd = pd_create_6_351();
if (!all_diagram_isotopies_test(pd,"pd_create_6_351")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_352...");
pd = pd_create_6_352();
if (!all_diagram_isotopies_test(pd,"pd_create_6_352")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_353...");
pd = pd_create_6_353();
if (!all_diagram_isotopies_test(pd,"pd_create_6_353")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_354...");
pd = pd_create_6_354();
if (!all_diagram_isotopies_test(pd,"pd_create_6_354")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_355...");
pd = pd_create_6_355();
if (!all_diagram_isotopies_test(pd,"pd_create_6_355")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_356...");
pd = pd_create_6_356();
if (!all_diagram_isotopies_test(pd,"pd_create_6_356")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_357...");
pd = pd_create_6_357();
if (!all_diagram_isotopies_test(pd,"pd_create_6_357")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_358...");
pd = pd_create_6_358();
if (!all_diagram_isotopies_test(pd,"pd_create_6_358")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_359...");
pd = pd_create_6_359();
if (!all_diagram_isotopies_test(pd,"pd_create_6_359")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_36...");
pd = pd_create_6_36();
if (!all_diagram_isotopies_test(pd,"pd_create_6_36")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_360...");
pd = pd_create_6_360();
if (!all_diagram_isotopies_test(pd,"pd_create_6_360")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_361...");
pd = pd_create_6_361();
if (!all_diagram_isotopies_test(pd,"pd_create_6_361")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_362...");
pd = pd_create_6_362();
if (!all_diagram_isotopies_test(pd,"pd_create_6_362")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_363...");
pd = pd_create_6_363();
if (!all_diagram_isotopies_test(pd,"pd_create_6_363")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_364...");
pd = pd_create_6_364();
if (!all_diagram_isotopies_test(pd,"pd_create_6_364")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_365...");
pd = pd_create_6_365();
if (!all_diagram_isotopies_test(pd,"pd_create_6_365")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_366...");
pd = pd_create_6_366();
if (!all_diagram_isotopies_test(pd,"pd_create_6_366")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_367...");
pd = pd_create_6_367();
if (!all_diagram_isotopies_test(pd,"pd_create_6_367")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_368...");
pd = pd_create_6_368();
if (!all_diagram_isotopies_test(pd,"pd_create_6_368")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_369...");
pd = pd_create_6_369();
if (!all_diagram_isotopies_test(pd,"pd_create_6_369")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_37...");
pd = pd_create_6_37();
if (!all_diagram_isotopies_test(pd,"pd_create_6_37")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_370...");
pd = pd_create_6_370();
if (!all_diagram_isotopies_test(pd,"pd_create_6_370")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_371...");
pd = pd_create_6_371();
if (!all_diagram_isotopies_test(pd,"pd_create_6_371")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_372...");
pd = pd_create_6_372();
if (!all_diagram_isotopies_test(pd,"pd_create_6_372")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_373...");
pd = pd_create_6_373();
if (!all_diagram_isotopies_test(pd,"pd_create_6_373")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_374...");
pd = pd_create_6_374();
if (!all_diagram_isotopies_test(pd,"pd_create_6_374")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_375...");
pd = pd_create_6_375();
if (!all_diagram_isotopies_test(pd,"pd_create_6_375")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_376...");
pd = pd_create_6_376();
if (!all_diagram_isotopies_test(pd,"pd_create_6_376")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_377...");
pd = pd_create_6_377();
if (!all_diagram_isotopies_test(pd,"pd_create_6_377")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_378...");
pd = pd_create_6_378();
if (!all_diagram_isotopies_test(pd,"pd_create_6_378")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_379...");
pd = pd_create_6_379();
if (!all_diagram_isotopies_test(pd,"pd_create_6_379")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_38...");
pd = pd_create_6_38();
if (!all_diagram_isotopies_test(pd,"pd_create_6_38")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_380...");
pd = pd_create_6_380();
if (!all_diagram_isotopies_test(pd,"pd_create_6_380")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_381...");
pd = pd_create_6_381();
if (!all_diagram_isotopies_test(pd,"pd_create_6_381")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_382...");
pd = pd_create_6_382();
if (!all_diagram_isotopies_test(pd,"pd_create_6_382")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_383...");
pd = pd_create_6_383();
if (!all_diagram_isotopies_test(pd,"pd_create_6_383")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_384...");
pd = pd_create_6_384();
if (!all_diagram_isotopies_test(pd,"pd_create_6_384")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_385...");
pd = pd_create_6_385();
if (!all_diagram_isotopies_test(pd,"pd_create_6_385")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_386...");
pd = pd_create_6_386();
if (!all_diagram_isotopies_test(pd,"pd_create_6_386")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_387...");
pd = pd_create_6_387();
if (!all_diagram_isotopies_test(pd,"pd_create_6_387")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_388...");
pd = pd_create_6_388();
if (!all_diagram_isotopies_test(pd,"pd_create_6_388")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_389...");
pd = pd_create_6_389();
if (!all_diagram_isotopies_test(pd,"pd_create_6_389")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_39...");
pd = pd_create_6_39();
if (!all_diagram_isotopies_test(pd,"pd_create_6_39")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_390...");
pd = pd_create_6_390();
if (!all_diagram_isotopies_test(pd,"pd_create_6_390")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_391...");
pd = pd_create_6_391();
if (!all_diagram_isotopies_test(pd,"pd_create_6_391")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_392...");
pd = pd_create_6_392();
if (!all_diagram_isotopies_test(pd,"pd_create_6_392")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_393...");
pd = pd_create_6_393();
if (!all_diagram_isotopies_test(pd,"pd_create_6_393")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_394...");
pd = pd_create_6_394();
if (!all_diagram_isotopies_test(pd,"pd_create_6_394")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_395...");
pd = pd_create_6_395();
if (!all_diagram_isotopies_test(pd,"pd_create_6_395")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_396...");
pd = pd_create_6_396();
if (!all_diagram_isotopies_test(pd,"pd_create_6_396")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_397...");
pd = pd_create_6_397();
if (!all_diagram_isotopies_test(pd,"pd_create_6_397")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_398...");
pd = pd_create_6_398();
if (!all_diagram_isotopies_test(pd,"pd_create_6_398")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_399...");
pd = pd_create_6_399();
if (!all_diagram_isotopies_test(pd,"pd_create_6_399")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_4...");
pd = pd_create_6_4();
if (!all_diagram_isotopies_test(pd,"pd_create_6_4")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_40...");
pd = pd_create_6_40();
if (!all_diagram_isotopies_test(pd,"pd_create_6_40")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_400...");
pd = pd_create_6_400();
if (!all_diagram_isotopies_test(pd,"pd_create_6_400")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_401...");
pd = pd_create_6_401();
if (!all_diagram_isotopies_test(pd,"pd_create_6_401")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_402...");
pd = pd_create_6_402();
if (!all_diagram_isotopies_test(pd,"pd_create_6_402")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_403...");
pd = pd_create_6_403();
if (!all_diagram_isotopies_test(pd,"pd_create_6_403")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_404...");
pd = pd_create_6_404();
if (!all_diagram_isotopies_test(pd,"pd_create_6_404")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_405...");
pd = pd_create_6_405();
if (!all_diagram_isotopies_test(pd,"pd_create_6_405")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_406...");
pd = pd_create_6_406();
if (!all_diagram_isotopies_test(pd,"pd_create_6_406")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_407...");
pd = pd_create_6_407();
if (!all_diagram_isotopies_test(pd,"pd_create_6_407")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_408...");
pd = pd_create_6_408();
if (!all_diagram_isotopies_test(pd,"pd_create_6_408")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_409...");
pd = pd_create_6_409();
if (!all_diagram_isotopies_test(pd,"pd_create_6_409")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_41...");
pd = pd_create_6_41();
if (!all_diagram_isotopies_test(pd,"pd_create_6_41")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_410...");
pd = pd_create_6_410();
if (!all_diagram_isotopies_test(pd,"pd_create_6_410")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_411...");
pd = pd_create_6_411();
if (!all_diagram_isotopies_test(pd,"pd_create_6_411")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_412...");
pd = pd_create_6_412();
if (!all_diagram_isotopies_test(pd,"pd_create_6_412")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_413...");
pd = pd_create_6_413();
if (!all_diagram_isotopies_test(pd,"pd_create_6_413")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_414...");
pd = pd_create_6_414();
if (!all_diagram_isotopies_test(pd,"pd_create_6_414")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_415...");
pd = pd_create_6_415();
if (!all_diagram_isotopies_test(pd,"pd_create_6_415")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_416...");
pd = pd_create_6_416();
if (!all_diagram_isotopies_test(pd,"pd_create_6_416")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_417...");
pd = pd_create_6_417();
if (!all_diagram_isotopies_test(pd,"pd_create_6_417")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_418...");
pd = pd_create_6_418();
if (!all_diagram_isotopies_test(pd,"pd_create_6_418")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_419...");
pd = pd_create_6_419();
if (!all_diagram_isotopies_test(pd,"pd_create_6_419")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_42...");
pd = pd_create_6_42();
if (!all_diagram_isotopies_test(pd,"pd_create_6_42")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_420...");
pd = pd_create_6_420();
if (!all_diagram_isotopies_test(pd,"pd_create_6_420")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_421...");
pd = pd_create_6_421();
if (!all_diagram_isotopies_test(pd,"pd_create_6_421")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_422...");
pd = pd_create_6_422();
if (!all_diagram_isotopies_test(pd,"pd_create_6_422")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_423...");
pd = pd_create_6_423();
if (!all_diagram_isotopies_test(pd,"pd_create_6_423")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_424...");
pd = pd_create_6_424();
if (!all_diagram_isotopies_test(pd,"pd_create_6_424")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_425...");
pd = pd_create_6_425();
if (!all_diagram_isotopies_test(pd,"pd_create_6_425")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_426...");
pd = pd_create_6_426();
if (!all_diagram_isotopies_test(pd,"pd_create_6_426")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_427...");
pd = pd_create_6_427();
if (!all_diagram_isotopies_test(pd,"pd_create_6_427")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_428...");
pd = pd_create_6_428();
if (!all_diagram_isotopies_test(pd,"pd_create_6_428")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_429...");
pd = pd_create_6_429();
if (!all_diagram_isotopies_test(pd,"pd_create_6_429")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_43...");
pd = pd_create_6_43();
if (!all_diagram_isotopies_test(pd,"pd_create_6_43")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_430...");
pd = pd_create_6_430();
if (!all_diagram_isotopies_test(pd,"pd_create_6_430")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_431...");
pd = pd_create_6_431();
if (!all_diagram_isotopies_test(pd,"pd_create_6_431")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_432...");
pd = pd_create_6_432();
if (!all_diagram_isotopies_test(pd,"pd_create_6_432")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_433...");
pd = pd_create_6_433();
if (!all_diagram_isotopies_test(pd,"pd_create_6_433")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_434...");
pd = pd_create_6_434();
if (!all_diagram_isotopies_test(pd,"pd_create_6_434")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_435...");
pd = pd_create_6_435();
if (!all_diagram_isotopies_test(pd,"pd_create_6_435")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_436...");
pd = pd_create_6_436();
if (!all_diagram_isotopies_test(pd,"pd_create_6_436")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_437...");
pd = pd_create_6_437();
if (!all_diagram_isotopies_test(pd,"pd_create_6_437")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_438...");
pd = pd_create_6_438();
if (!all_diagram_isotopies_test(pd,"pd_create_6_438")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_439...");
pd = pd_create_6_439();
if (!all_diagram_isotopies_test(pd,"pd_create_6_439")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_44...");
pd = pd_create_6_44();
if (!all_diagram_isotopies_test(pd,"pd_create_6_44")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_440...");
pd = pd_create_6_440();
if (!all_diagram_isotopies_test(pd,"pd_create_6_440")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_441...");
pd = pd_create_6_441();
if (!all_diagram_isotopies_test(pd,"pd_create_6_441")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_442...");
pd = pd_create_6_442();
if (!all_diagram_isotopies_test(pd,"pd_create_6_442")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_443...");
pd = pd_create_6_443();
if (!all_diagram_isotopies_test(pd,"pd_create_6_443")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_444...");
pd = pd_create_6_444();
if (!all_diagram_isotopies_test(pd,"pd_create_6_444")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_445...");
pd = pd_create_6_445();
if (!all_diagram_isotopies_test(pd,"pd_create_6_445")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_446...");
pd = pd_create_6_446();
if (!all_diagram_isotopies_test(pd,"pd_create_6_446")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_447...");
pd = pd_create_6_447();
if (!all_diagram_isotopies_test(pd,"pd_create_6_447")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_448...");
pd = pd_create_6_448();
if (!all_diagram_isotopies_test(pd,"pd_create_6_448")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_449...");
pd = pd_create_6_449();
if (!all_diagram_isotopies_test(pd,"pd_create_6_449")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_45...");
pd = pd_create_6_45();
if (!all_diagram_isotopies_test(pd,"pd_create_6_45")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_450...");
pd = pd_create_6_450();
if (!all_diagram_isotopies_test(pd,"pd_create_6_450")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_451...");
pd = pd_create_6_451();
if (!all_diagram_isotopies_test(pd,"pd_create_6_451")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_452...");
pd = pd_create_6_452();
if (!all_diagram_isotopies_test(pd,"pd_create_6_452")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_453...");
pd = pd_create_6_453();
if (!all_diagram_isotopies_test(pd,"pd_create_6_453")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_454...");
pd = pd_create_6_454();
if (!all_diagram_isotopies_test(pd,"pd_create_6_454")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_455...");
pd = pd_create_6_455();
if (!all_diagram_isotopies_test(pd,"pd_create_6_455")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_456...");
pd = pd_create_6_456();
if (!all_diagram_isotopies_test(pd,"pd_create_6_456")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_457...");
pd = pd_create_6_457();
if (!all_diagram_isotopies_test(pd,"pd_create_6_457")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_458...");
pd = pd_create_6_458();
if (!all_diagram_isotopies_test(pd,"pd_create_6_458")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_459...");
pd = pd_create_6_459();
if (!all_diagram_isotopies_test(pd,"pd_create_6_459")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_46...");
pd = pd_create_6_46();
if (!all_diagram_isotopies_test(pd,"pd_create_6_46")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_460...");
pd = pd_create_6_460();
if (!all_diagram_isotopies_test(pd,"pd_create_6_460")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_461...");
pd = pd_create_6_461();
if (!all_diagram_isotopies_test(pd,"pd_create_6_461")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_462...");
pd = pd_create_6_462();
if (!all_diagram_isotopies_test(pd,"pd_create_6_462")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_463...");
pd = pd_create_6_463();
if (!all_diagram_isotopies_test(pd,"pd_create_6_463")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_464...");
pd = pd_create_6_464();
if (!all_diagram_isotopies_test(pd,"pd_create_6_464")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_465...");
pd = pd_create_6_465();
if (!all_diagram_isotopies_test(pd,"pd_create_6_465")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_466...");
pd = pd_create_6_466();
if (!all_diagram_isotopies_test(pd,"pd_create_6_466")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_467...");
pd = pd_create_6_467();
if (!all_diagram_isotopies_test(pd,"pd_create_6_467")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_468...");
pd = pd_create_6_468();
if (!all_diagram_isotopies_test(pd,"pd_create_6_468")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_469...");
pd = pd_create_6_469();
if (!all_diagram_isotopies_test(pd,"pd_create_6_469")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_47...");
pd = pd_create_6_47();
if (!all_diagram_isotopies_test(pd,"pd_create_6_47")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_470...");
pd = pd_create_6_470();
if (!all_diagram_isotopies_test(pd,"pd_create_6_470")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_471...");
pd = pd_create_6_471();
if (!all_diagram_isotopies_test(pd,"pd_create_6_471")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_472...");
pd = pd_create_6_472();
if (!all_diagram_isotopies_test(pd,"pd_create_6_472")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_473...");
pd = pd_create_6_473();
if (!all_diagram_isotopies_test(pd,"pd_create_6_473")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_474...");
pd = pd_create_6_474();
if (!all_diagram_isotopies_test(pd,"pd_create_6_474")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_475...");
pd = pd_create_6_475();
if (!all_diagram_isotopies_test(pd,"pd_create_6_475")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_476...");
pd = pd_create_6_476();
if (!all_diagram_isotopies_test(pd,"pd_create_6_476")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_477...");
pd = pd_create_6_477();
if (!all_diagram_isotopies_test(pd,"pd_create_6_477")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_478...");
pd = pd_create_6_478();
if (!all_diagram_isotopies_test(pd,"pd_create_6_478")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_479...");
pd = pd_create_6_479();
if (!all_diagram_isotopies_test(pd,"pd_create_6_479")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_48...");
pd = pd_create_6_48();
if (!all_diagram_isotopies_test(pd,"pd_create_6_48")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_480...");
pd = pd_create_6_480();
if (!all_diagram_isotopies_test(pd,"pd_create_6_480")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_481...");
pd = pd_create_6_481();
if (!all_diagram_isotopies_test(pd,"pd_create_6_481")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_482...");
pd = pd_create_6_482();
if (!all_diagram_isotopies_test(pd,"pd_create_6_482")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_483...");
pd = pd_create_6_483();
if (!all_diagram_isotopies_test(pd,"pd_create_6_483")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_484...");
pd = pd_create_6_484();
if (!all_diagram_isotopies_test(pd,"pd_create_6_484")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_485...");
pd = pd_create_6_485();
if (!all_diagram_isotopies_test(pd,"pd_create_6_485")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_486...");
pd = pd_create_6_486();
if (!all_diagram_isotopies_test(pd,"pd_create_6_486")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_487...");
pd = pd_create_6_487();
if (!all_diagram_isotopies_test(pd,"pd_create_6_487")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_488...");
pd = pd_create_6_488();
if (!all_diagram_isotopies_test(pd,"pd_create_6_488")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_489...");
pd = pd_create_6_489();
if (!all_diagram_isotopies_test(pd,"pd_create_6_489")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_49...");
pd = pd_create_6_49();
if (!all_diagram_isotopies_test(pd,"pd_create_6_49")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_490...");
pd = pd_create_6_490();
if (!all_diagram_isotopies_test(pd,"pd_create_6_490")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_491...");
pd = pd_create_6_491();
if (!all_diagram_isotopies_test(pd,"pd_create_6_491")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_492...");
pd = pd_create_6_492();
if (!all_diagram_isotopies_test(pd,"pd_create_6_492")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_493...");
pd = pd_create_6_493();
if (!all_diagram_isotopies_test(pd,"pd_create_6_493")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_494...");
pd = pd_create_6_494();
if (!all_diagram_isotopies_test(pd,"pd_create_6_494")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_495...");
pd = pd_create_6_495();
if (!all_diagram_isotopies_test(pd,"pd_create_6_495")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_496...");
pd = pd_create_6_496();
if (!all_diagram_isotopies_test(pd,"pd_create_6_496")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_497...");
pd = pd_create_6_497();
if (!all_diagram_isotopies_test(pd,"pd_create_6_497")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_498...");
pd = pd_create_6_498();
if (!all_diagram_isotopies_test(pd,"pd_create_6_498")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_499...");
pd = pd_create_6_499();
if (!all_diagram_isotopies_test(pd,"pd_create_6_499")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_5...");
pd = pd_create_6_5();
if (!all_diagram_isotopies_test(pd,"pd_create_6_5")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_50...");
pd = pd_create_6_50();
if (!all_diagram_isotopies_test(pd,"pd_create_6_50")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_500...");
pd = pd_create_6_500();
if (!all_diagram_isotopies_test(pd,"pd_create_6_500")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_501...");
pd = pd_create_6_501();
if (!all_diagram_isotopies_test(pd,"pd_create_6_501")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_502...");
pd = pd_create_6_502();
if (!all_diagram_isotopies_test(pd,"pd_create_6_502")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_503...");
pd = pd_create_6_503();
if (!all_diagram_isotopies_test(pd,"pd_create_6_503")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_504...");
pd = pd_create_6_504();
if (!all_diagram_isotopies_test(pd,"pd_create_6_504")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_505...");
pd = pd_create_6_505();
if (!all_diagram_isotopies_test(pd,"pd_create_6_505")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_506...");
pd = pd_create_6_506();
if (!all_diagram_isotopies_test(pd,"pd_create_6_506")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_507...");
pd = pd_create_6_507();
if (!all_diagram_isotopies_test(pd,"pd_create_6_507")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_508...");
pd = pd_create_6_508();
if (!all_diagram_isotopies_test(pd,"pd_create_6_508")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_509...");
pd = pd_create_6_509();
if (!all_diagram_isotopies_test(pd,"pd_create_6_509")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_51...");
pd = pd_create_6_51();
if (!all_diagram_isotopies_test(pd,"pd_create_6_51")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_510...");
pd = pd_create_6_510();
if (!all_diagram_isotopies_test(pd,"pd_create_6_510")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_511...");
pd = pd_create_6_511();
if (!all_diagram_isotopies_test(pd,"pd_create_6_511")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_512...");
pd = pd_create_6_512();
if (!all_diagram_isotopies_test(pd,"pd_create_6_512")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_513...");
pd = pd_create_6_513();
if (!all_diagram_isotopies_test(pd,"pd_create_6_513")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_514...");
pd = pd_create_6_514();
if (!all_diagram_isotopies_test(pd,"pd_create_6_514")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_515...");
pd = pd_create_6_515();
if (!all_diagram_isotopies_test(pd,"pd_create_6_515")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_516...");
pd = pd_create_6_516();
if (!all_diagram_isotopies_test(pd,"pd_create_6_516")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_517...");
pd = pd_create_6_517();
if (!all_diagram_isotopies_test(pd,"pd_create_6_517")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_518...");
pd = pd_create_6_518();
if (!all_diagram_isotopies_test(pd,"pd_create_6_518")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_519...");
pd = pd_create_6_519();
if (!all_diagram_isotopies_test(pd,"pd_create_6_519")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_52...");
pd = pd_create_6_52();
if (!all_diagram_isotopies_test(pd,"pd_create_6_52")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_520...");
pd = pd_create_6_520();
if (!all_diagram_isotopies_test(pd,"pd_create_6_520")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_521...");
pd = pd_create_6_521();
if (!all_diagram_isotopies_test(pd,"pd_create_6_521")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_522...");
pd = pd_create_6_522();
if (!all_diagram_isotopies_test(pd,"pd_create_6_522")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_523...");
pd = pd_create_6_523();
if (!all_diagram_isotopies_test(pd,"pd_create_6_523")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_524...");
pd = pd_create_6_524();
if (!all_diagram_isotopies_test(pd,"pd_create_6_524")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_525...");
pd = pd_create_6_525();
if (!all_diagram_isotopies_test(pd,"pd_create_6_525")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_526...");
pd = pd_create_6_526();
if (!all_diagram_isotopies_test(pd,"pd_create_6_526")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_527...");
pd = pd_create_6_527();
if (!all_diagram_isotopies_test(pd,"pd_create_6_527")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_528...");
pd = pd_create_6_528();
if (!all_diagram_isotopies_test(pd,"pd_create_6_528")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_529...");
pd = pd_create_6_529();
if (!all_diagram_isotopies_test(pd,"pd_create_6_529")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_53...");
pd = pd_create_6_53();
if (!all_diagram_isotopies_test(pd,"pd_create_6_53")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_530...");
pd = pd_create_6_530();
if (!all_diagram_isotopies_test(pd,"pd_create_6_530")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_531...");
pd = pd_create_6_531();
if (!all_diagram_isotopies_test(pd,"pd_create_6_531")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_532...");
pd = pd_create_6_532();
if (!all_diagram_isotopies_test(pd,"pd_create_6_532")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_533...");
pd = pd_create_6_533();
if (!all_diagram_isotopies_test(pd,"pd_create_6_533")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_534...");
pd = pd_create_6_534();
if (!all_diagram_isotopies_test(pd,"pd_create_6_534")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_535...");
pd = pd_create_6_535();
if (!all_diagram_isotopies_test(pd,"pd_create_6_535")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_536...");
pd = pd_create_6_536();
if (!all_diagram_isotopies_test(pd,"pd_create_6_536")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_537...");
pd = pd_create_6_537();
if (!all_diagram_isotopies_test(pd,"pd_create_6_537")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_538...");
pd = pd_create_6_538();
if (!all_diagram_isotopies_test(pd,"pd_create_6_538")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_539...");
pd = pd_create_6_539();
if (!all_diagram_isotopies_test(pd,"pd_create_6_539")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_54...");
pd = pd_create_6_54();
if (!all_diagram_isotopies_test(pd,"pd_create_6_54")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_540...");
pd = pd_create_6_540();
if (!all_diagram_isotopies_test(pd,"pd_create_6_540")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_541...");
pd = pd_create_6_541();
if (!all_diagram_isotopies_test(pd,"pd_create_6_541")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_542...");
pd = pd_create_6_542();
if (!all_diagram_isotopies_test(pd,"pd_create_6_542")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_543...");
pd = pd_create_6_543();
if (!all_diagram_isotopies_test(pd,"pd_create_6_543")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_544...");
pd = pd_create_6_544();
if (!all_diagram_isotopies_test(pd,"pd_create_6_544")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_545...");
pd = pd_create_6_545();
if (!all_diagram_isotopies_test(pd,"pd_create_6_545")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_546...");
pd = pd_create_6_546();
if (!all_diagram_isotopies_test(pd,"pd_create_6_546")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_547...");
pd = pd_create_6_547();
if (!all_diagram_isotopies_test(pd,"pd_create_6_547")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_548...");
pd = pd_create_6_548();
if (!all_diagram_isotopies_test(pd,"pd_create_6_548")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_549...");
pd = pd_create_6_549();
if (!all_diagram_isotopies_test(pd,"pd_create_6_549")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_55...");
pd = pd_create_6_55();
if (!all_diagram_isotopies_test(pd,"pd_create_6_55")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_550...");
pd = pd_create_6_550();
if (!all_diagram_isotopies_test(pd,"pd_create_6_550")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_551...");
pd = pd_create_6_551();
if (!all_diagram_isotopies_test(pd,"pd_create_6_551")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_552...");
pd = pd_create_6_552();
if (!all_diagram_isotopies_test(pd,"pd_create_6_552")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_553...");
pd = pd_create_6_553();
if (!all_diagram_isotopies_test(pd,"pd_create_6_553")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_554...");
pd = pd_create_6_554();
if (!all_diagram_isotopies_test(pd,"pd_create_6_554")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_555...");
pd = pd_create_6_555();
if (!all_diagram_isotopies_test(pd,"pd_create_6_555")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_556...");
pd = pd_create_6_556();
if (!all_diagram_isotopies_test(pd,"pd_create_6_556")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_557...");
pd = pd_create_6_557();
if (!all_diagram_isotopies_test(pd,"pd_create_6_557")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_558...");
pd = pd_create_6_558();
if (!all_diagram_isotopies_test(pd,"pd_create_6_558")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_559...");
pd = pd_create_6_559();
if (!all_diagram_isotopies_test(pd,"pd_create_6_559")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_56...");
pd = pd_create_6_56();
if (!all_diagram_isotopies_test(pd,"pd_create_6_56")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_560...");
pd = pd_create_6_560();
if (!all_diagram_isotopies_test(pd,"pd_create_6_560")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_561...");
pd = pd_create_6_561();
if (!all_diagram_isotopies_test(pd,"pd_create_6_561")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_562...");
pd = pd_create_6_562();
if (!all_diagram_isotopies_test(pd,"pd_create_6_562")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_563...");
pd = pd_create_6_563();
if (!all_diagram_isotopies_test(pd,"pd_create_6_563")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_564...");
pd = pd_create_6_564();
if (!all_diagram_isotopies_test(pd,"pd_create_6_564")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_565...");
pd = pd_create_6_565();
if (!all_diagram_isotopies_test(pd,"pd_create_6_565")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_566...");
pd = pd_create_6_566();
if (!all_diagram_isotopies_test(pd,"pd_create_6_566")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_567...");
pd = pd_create_6_567();
if (!all_diagram_isotopies_test(pd,"pd_create_6_567")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_568...");
pd = pd_create_6_568();
if (!all_diagram_isotopies_test(pd,"pd_create_6_568")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_569...");
pd = pd_create_6_569();
if (!all_diagram_isotopies_test(pd,"pd_create_6_569")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_57...");
pd = pd_create_6_57();
if (!all_diagram_isotopies_test(pd,"pd_create_6_57")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_570...");
pd = pd_create_6_570();
if (!all_diagram_isotopies_test(pd,"pd_create_6_570")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_571...");
pd = pd_create_6_571();
if (!all_diagram_isotopies_test(pd,"pd_create_6_571")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_572...");
pd = pd_create_6_572();
if (!all_diagram_isotopies_test(pd,"pd_create_6_572")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_573...");
pd = pd_create_6_573();
if (!all_diagram_isotopies_test(pd,"pd_create_6_573")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_574...");
pd = pd_create_6_574();
if (!all_diagram_isotopies_test(pd,"pd_create_6_574")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_575...");
pd = pd_create_6_575();
if (!all_diagram_isotopies_test(pd,"pd_create_6_575")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_576...");
pd = pd_create_6_576();
if (!all_diagram_isotopies_test(pd,"pd_create_6_576")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_577...");
pd = pd_create_6_577();
if (!all_diagram_isotopies_test(pd,"pd_create_6_577")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_578...");
pd = pd_create_6_578();
if (!all_diagram_isotopies_test(pd,"pd_create_6_578")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_579...");
pd = pd_create_6_579();
if (!all_diagram_isotopies_test(pd,"pd_create_6_579")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_58...");
pd = pd_create_6_58();
if (!all_diagram_isotopies_test(pd,"pd_create_6_58")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_580...");
pd = pd_create_6_580();
if (!all_diagram_isotopies_test(pd,"pd_create_6_580")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_581...");
pd = pd_create_6_581();
if (!all_diagram_isotopies_test(pd,"pd_create_6_581")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_582...");
pd = pd_create_6_582();
if (!all_diagram_isotopies_test(pd,"pd_create_6_582")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_583...");
pd = pd_create_6_583();
if (!all_diagram_isotopies_test(pd,"pd_create_6_583")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_584...");
pd = pd_create_6_584();
if (!all_diagram_isotopies_test(pd,"pd_create_6_584")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_585...");
pd = pd_create_6_585();
if (!all_diagram_isotopies_test(pd,"pd_create_6_585")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_586...");
pd = pd_create_6_586();
if (!all_diagram_isotopies_test(pd,"pd_create_6_586")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_587...");
pd = pd_create_6_587();
if (!all_diagram_isotopies_test(pd,"pd_create_6_587")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_588...");
pd = pd_create_6_588();
if (!all_diagram_isotopies_test(pd,"pd_create_6_588")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_589...");
pd = pd_create_6_589();
if (!all_diagram_isotopies_test(pd,"pd_create_6_589")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_59...");
pd = pd_create_6_59();
if (!all_diagram_isotopies_test(pd,"pd_create_6_59")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_590...");
pd = pd_create_6_590();
if (!all_diagram_isotopies_test(pd,"pd_create_6_590")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_591...");
pd = pd_create_6_591();
if (!all_diagram_isotopies_test(pd,"pd_create_6_591")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_592...");
pd = pd_create_6_592();
if (!all_diagram_isotopies_test(pd,"pd_create_6_592")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_593...");
pd = pd_create_6_593();
if (!all_diagram_isotopies_test(pd,"pd_create_6_593")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_594...");
pd = pd_create_6_594();
if (!all_diagram_isotopies_test(pd,"pd_create_6_594")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_595...");
pd = pd_create_6_595();
if (!all_diagram_isotopies_test(pd,"pd_create_6_595")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_596...");
pd = pd_create_6_596();
if (!all_diagram_isotopies_test(pd,"pd_create_6_596")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_597...");
pd = pd_create_6_597();
if (!all_diagram_isotopies_test(pd,"pd_create_6_597")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_598...");
pd = pd_create_6_598();
if (!all_diagram_isotopies_test(pd,"pd_create_6_598")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_599...");
pd = pd_create_6_599();
if (!all_diagram_isotopies_test(pd,"pd_create_6_599")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_6...");
pd = pd_create_6_6();
if (!all_diagram_isotopies_test(pd,"pd_create_6_6")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_60...");
pd = pd_create_6_60();
if (!all_diagram_isotopies_test(pd,"pd_create_6_60")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_600...");
pd = pd_create_6_600();
if (!all_diagram_isotopies_test(pd,"pd_create_6_600")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_601...");
pd = pd_create_6_601();
if (!all_diagram_isotopies_test(pd,"pd_create_6_601")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_602...");
pd = pd_create_6_602();
if (!all_diagram_isotopies_test(pd,"pd_create_6_602")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_603...");
pd = pd_create_6_603();
if (!all_diagram_isotopies_test(pd,"pd_create_6_603")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_604...");
pd = pd_create_6_604();
if (!all_diagram_isotopies_test(pd,"pd_create_6_604")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_605...");
pd = pd_create_6_605();
if (!all_diagram_isotopies_test(pd,"pd_create_6_605")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_606...");
pd = pd_create_6_606();
if (!all_diagram_isotopies_test(pd,"pd_create_6_606")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_607...");
pd = pd_create_6_607();
if (!all_diagram_isotopies_test(pd,"pd_create_6_607")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_608...");
pd = pd_create_6_608();
if (!all_diagram_isotopies_test(pd,"pd_create_6_608")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_609...");
pd = pd_create_6_609();
if (!all_diagram_isotopies_test(pd,"pd_create_6_609")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_61...");
pd = pd_create_6_61();
if (!all_diagram_isotopies_test(pd,"pd_create_6_61")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_610...");
pd = pd_create_6_610();
if (!all_diagram_isotopies_test(pd,"pd_create_6_610")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_611...");
pd = pd_create_6_611();
if (!all_diagram_isotopies_test(pd,"pd_create_6_611")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_612...");
pd = pd_create_6_612();
if (!all_diagram_isotopies_test(pd,"pd_create_6_612")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_613...");
pd = pd_create_6_613();
if (!all_diagram_isotopies_test(pd,"pd_create_6_613")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_614...");
pd = pd_create_6_614();
if (!all_diagram_isotopies_test(pd,"pd_create_6_614")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_615...");
pd = pd_create_6_615();
if (!all_diagram_isotopies_test(pd,"pd_create_6_615")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_616...");
pd = pd_create_6_616();
if (!all_diagram_isotopies_test(pd,"pd_create_6_616")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_617...");
pd = pd_create_6_617();
if (!all_diagram_isotopies_test(pd,"pd_create_6_617")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_618...");
pd = pd_create_6_618();
if (!all_diagram_isotopies_test(pd,"pd_create_6_618")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_619...");
pd = pd_create_6_619();
if (!all_diagram_isotopies_test(pd,"pd_create_6_619")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_62...");
pd = pd_create_6_62();
if (!all_diagram_isotopies_test(pd,"pd_create_6_62")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_620...");
pd = pd_create_6_620();
if (!all_diagram_isotopies_test(pd,"pd_create_6_620")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_621...");
pd = pd_create_6_621();
if (!all_diagram_isotopies_test(pd,"pd_create_6_621")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_622...");
pd = pd_create_6_622();
if (!all_diagram_isotopies_test(pd,"pd_create_6_622")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_623...");
pd = pd_create_6_623();
if (!all_diagram_isotopies_test(pd,"pd_create_6_623")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_624...");
pd = pd_create_6_624();
if (!all_diagram_isotopies_test(pd,"pd_create_6_624")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_625...");
pd = pd_create_6_625();
if (!all_diagram_isotopies_test(pd,"pd_create_6_625")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_626...");
pd = pd_create_6_626();
if (!all_diagram_isotopies_test(pd,"pd_create_6_626")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_627...");
pd = pd_create_6_627();
if (!all_diagram_isotopies_test(pd,"pd_create_6_627")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_628...");
pd = pd_create_6_628();
if (!all_diagram_isotopies_test(pd,"pd_create_6_628")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_629...");
pd = pd_create_6_629();
if (!all_diagram_isotopies_test(pd,"pd_create_6_629")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_63...");
pd = pd_create_6_63();
if (!all_diagram_isotopies_test(pd,"pd_create_6_63")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_630...");
pd = pd_create_6_630();
if (!all_diagram_isotopies_test(pd,"pd_create_6_630")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_631...");
pd = pd_create_6_631();
if (!all_diagram_isotopies_test(pd,"pd_create_6_631")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_632...");
pd = pd_create_6_632();
if (!all_diagram_isotopies_test(pd,"pd_create_6_632")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_633...");
pd = pd_create_6_633();
if (!all_diagram_isotopies_test(pd,"pd_create_6_633")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_634...");
pd = pd_create_6_634();
if (!all_diagram_isotopies_test(pd,"pd_create_6_634")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_635...");
pd = pd_create_6_635();
if (!all_diagram_isotopies_test(pd,"pd_create_6_635")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_636...");
pd = pd_create_6_636();
if (!all_diagram_isotopies_test(pd,"pd_create_6_636")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_637...");
pd = pd_create_6_637();
if (!all_diagram_isotopies_test(pd,"pd_create_6_637")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_638...");
pd = pd_create_6_638();
if (!all_diagram_isotopies_test(pd,"pd_create_6_638")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_639...");
pd = pd_create_6_639();
if (!all_diagram_isotopies_test(pd,"pd_create_6_639")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_64...");
pd = pd_create_6_64();
if (!all_diagram_isotopies_test(pd,"pd_create_6_64")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_640...");
pd = pd_create_6_640();
if (!all_diagram_isotopies_test(pd,"pd_create_6_640")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_641...");
pd = pd_create_6_641();
if (!all_diagram_isotopies_test(pd,"pd_create_6_641")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_642...");
pd = pd_create_6_642();
if (!all_diagram_isotopies_test(pd,"pd_create_6_642")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_643...");
pd = pd_create_6_643();
if (!all_diagram_isotopies_test(pd,"pd_create_6_643")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_644...");
pd = pd_create_6_644();
if (!all_diagram_isotopies_test(pd,"pd_create_6_644")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_645...");
pd = pd_create_6_645();
if (!all_diagram_isotopies_test(pd,"pd_create_6_645")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_646...");
pd = pd_create_6_646();
if (!all_diagram_isotopies_test(pd,"pd_create_6_646")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_647...");
pd = pd_create_6_647();
if (!all_diagram_isotopies_test(pd,"pd_create_6_647")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_648...");
pd = pd_create_6_648();
if (!all_diagram_isotopies_test(pd,"pd_create_6_648")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_649...");
pd = pd_create_6_649();
if (!all_diagram_isotopies_test(pd,"pd_create_6_649")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_65...");
pd = pd_create_6_65();
if (!all_diagram_isotopies_test(pd,"pd_create_6_65")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_650...");
pd = pd_create_6_650();
if (!all_diagram_isotopies_test(pd,"pd_create_6_650")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_651...");
pd = pd_create_6_651();
if (!all_diagram_isotopies_test(pd,"pd_create_6_651")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_652...");
pd = pd_create_6_652();
if (!all_diagram_isotopies_test(pd,"pd_create_6_652")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_653...");
pd = pd_create_6_653();
if (!all_diagram_isotopies_test(pd,"pd_create_6_653")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_654...");
pd = pd_create_6_654();
if (!all_diagram_isotopies_test(pd,"pd_create_6_654")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_655...");
pd = pd_create_6_655();
if (!all_diagram_isotopies_test(pd,"pd_create_6_655")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_656...");
pd = pd_create_6_656();
if (!all_diagram_isotopies_test(pd,"pd_create_6_656")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_657...");
pd = pd_create_6_657();
if (!all_diagram_isotopies_test(pd,"pd_create_6_657")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_658...");
pd = pd_create_6_658();
if (!all_diagram_isotopies_test(pd,"pd_create_6_658")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_659...");
pd = pd_create_6_659();
if (!all_diagram_isotopies_test(pd,"pd_create_6_659")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_66...");
pd = pd_create_6_66();
if (!all_diagram_isotopies_test(pd,"pd_create_6_66")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_660...");
pd = pd_create_6_660();
if (!all_diagram_isotopies_test(pd,"pd_create_6_660")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_661...");
pd = pd_create_6_661();
if (!all_diagram_isotopies_test(pd,"pd_create_6_661")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_662...");
pd = pd_create_6_662();
if (!all_diagram_isotopies_test(pd,"pd_create_6_662")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_663...");
pd = pd_create_6_663();
if (!all_diagram_isotopies_test(pd,"pd_create_6_663")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_664...");
pd = pd_create_6_664();
if (!all_diagram_isotopies_test(pd,"pd_create_6_664")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_665...");
pd = pd_create_6_665();
if (!all_diagram_isotopies_test(pd,"pd_create_6_665")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_666...");
pd = pd_create_6_666();
if (!all_diagram_isotopies_test(pd,"pd_create_6_666")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_667...");
pd = pd_create_6_667();
if (!all_diagram_isotopies_test(pd,"pd_create_6_667")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_668...");
pd = pd_create_6_668();
if (!all_diagram_isotopies_test(pd,"pd_create_6_668")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_669...");
pd = pd_create_6_669();
if (!all_diagram_isotopies_test(pd,"pd_create_6_669")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_67...");
pd = pd_create_6_67();
if (!all_diagram_isotopies_test(pd,"pd_create_6_67")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_670...");
pd = pd_create_6_670();
if (!all_diagram_isotopies_test(pd,"pd_create_6_670")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_671...");
pd = pd_create_6_671();
if (!all_diagram_isotopies_test(pd,"pd_create_6_671")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_672...");
pd = pd_create_6_672();
if (!all_diagram_isotopies_test(pd,"pd_create_6_672")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_673...");
pd = pd_create_6_673();
if (!all_diagram_isotopies_test(pd,"pd_create_6_673")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_674...");
pd = pd_create_6_674();
if (!all_diagram_isotopies_test(pd,"pd_create_6_674")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_675...");
pd = pd_create_6_675();
if (!all_diagram_isotopies_test(pd,"pd_create_6_675")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_676...");
pd = pd_create_6_676();
if (!all_diagram_isotopies_test(pd,"pd_create_6_676")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_677...");
pd = pd_create_6_677();
if (!all_diagram_isotopies_test(pd,"pd_create_6_677")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_678...");
pd = pd_create_6_678();
if (!all_diagram_isotopies_test(pd,"pd_create_6_678")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_679...");
pd = pd_create_6_679();
if (!all_diagram_isotopies_test(pd,"pd_create_6_679")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_68...");
pd = pd_create_6_68();
if (!all_diagram_isotopies_test(pd,"pd_create_6_68")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_680...");
pd = pd_create_6_680();
if (!all_diagram_isotopies_test(pd,"pd_create_6_680")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_681...");
pd = pd_create_6_681();
if (!all_diagram_isotopies_test(pd,"pd_create_6_681")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_682...");
pd = pd_create_6_682();
if (!all_diagram_isotopies_test(pd,"pd_create_6_682")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_683...");
pd = pd_create_6_683();
if (!all_diagram_isotopies_test(pd,"pd_create_6_683")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_684...");
pd = pd_create_6_684();
if (!all_diagram_isotopies_test(pd,"pd_create_6_684")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_685...");
pd = pd_create_6_685();
if (!all_diagram_isotopies_test(pd,"pd_create_6_685")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_686...");
pd = pd_create_6_686();
if (!all_diagram_isotopies_test(pd,"pd_create_6_686")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_687...");
pd = pd_create_6_687();
if (!all_diagram_isotopies_test(pd,"pd_create_6_687")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_688...");
pd = pd_create_6_688();
if (!all_diagram_isotopies_test(pd,"pd_create_6_688")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_689...");
pd = pd_create_6_689();
if (!all_diagram_isotopies_test(pd,"pd_create_6_689")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_69...");
pd = pd_create_6_69();
if (!all_diagram_isotopies_test(pd,"pd_create_6_69")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_690...");
pd = pd_create_6_690();
if (!all_diagram_isotopies_test(pd,"pd_create_6_690")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_691...");
pd = pd_create_6_691();
if (!all_diagram_isotopies_test(pd,"pd_create_6_691")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_692...");
pd = pd_create_6_692();
if (!all_diagram_isotopies_test(pd,"pd_create_6_692")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_693...");
pd = pd_create_6_693();
if (!all_diagram_isotopies_test(pd,"pd_create_6_693")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_694...");
pd = pd_create_6_694();
if (!all_diagram_isotopies_test(pd,"pd_create_6_694")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_695...");
pd = pd_create_6_695();
if (!all_diagram_isotopies_test(pd,"pd_create_6_695")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_696...");
pd = pd_create_6_696();
if (!all_diagram_isotopies_test(pd,"pd_create_6_696")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_697...");
pd = pd_create_6_697();
if (!all_diagram_isotopies_test(pd,"pd_create_6_697")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_698...");
pd = pd_create_6_698();
if (!all_diagram_isotopies_test(pd,"pd_create_6_698")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_699...");
pd = pd_create_6_699();
if (!all_diagram_isotopies_test(pd,"pd_create_6_699")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_7...");
pd = pd_create_6_7();
if (!all_diagram_isotopies_test(pd,"pd_create_6_7")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_70...");
pd = pd_create_6_70();
if (!all_diagram_isotopies_test(pd,"pd_create_6_70")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_700...");
pd = pd_create_6_700();
if (!all_diagram_isotopies_test(pd,"pd_create_6_700")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_701...");
pd = pd_create_6_701();
if (!all_diagram_isotopies_test(pd,"pd_create_6_701")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_702...");
pd = pd_create_6_702();
if (!all_diagram_isotopies_test(pd,"pd_create_6_702")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_703...");
pd = pd_create_6_703();
if (!all_diagram_isotopies_test(pd,"pd_create_6_703")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_704...");
pd = pd_create_6_704();
if (!all_diagram_isotopies_test(pd,"pd_create_6_704")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_705...");
pd = pd_create_6_705();
if (!all_diagram_isotopies_test(pd,"pd_create_6_705")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_706...");
pd = pd_create_6_706();
if (!all_diagram_isotopies_test(pd,"pd_create_6_706")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_707...");
pd = pd_create_6_707();
if (!all_diagram_isotopies_test(pd,"pd_create_6_707")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_708...");
pd = pd_create_6_708();
if (!all_diagram_isotopies_test(pd,"pd_create_6_708")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_709...");
pd = pd_create_6_709();
if (!all_diagram_isotopies_test(pd,"pd_create_6_709")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_71...");
pd = pd_create_6_71();
if (!all_diagram_isotopies_test(pd,"pd_create_6_71")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_710...");
pd = pd_create_6_710();
if (!all_diagram_isotopies_test(pd,"pd_create_6_710")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_711...");
pd = pd_create_6_711();
if (!all_diagram_isotopies_test(pd,"pd_create_6_711")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_712...");
pd = pd_create_6_712();
if (!all_diagram_isotopies_test(pd,"pd_create_6_712")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_713...");
pd = pd_create_6_713();
if (!all_diagram_isotopies_test(pd,"pd_create_6_713")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_714...");
pd = pd_create_6_714();
if (!all_diagram_isotopies_test(pd,"pd_create_6_714")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_715...");
pd = pd_create_6_715();
if (!all_diagram_isotopies_test(pd,"pd_create_6_715")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_716...");
pd = pd_create_6_716();
if (!all_diagram_isotopies_test(pd,"pd_create_6_716")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_717...");
pd = pd_create_6_717();
if (!all_diagram_isotopies_test(pd,"pd_create_6_717")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_718...");
pd = pd_create_6_718();
if (!all_diagram_isotopies_test(pd,"pd_create_6_718")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_719...");
pd = pd_create_6_719();
if (!all_diagram_isotopies_test(pd,"pd_create_6_719")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_72...");
pd = pd_create_6_72();
if (!all_diagram_isotopies_test(pd,"pd_create_6_72")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_720...");
pd = pd_create_6_720();
if (!all_diagram_isotopies_test(pd,"pd_create_6_720")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_721...");
pd = pd_create_6_721();
if (!all_diagram_isotopies_test(pd,"pd_create_6_721")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_722...");
pd = pd_create_6_722();
if (!all_diagram_isotopies_test(pd,"pd_create_6_722")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_723...");
pd = pd_create_6_723();
if (!all_diagram_isotopies_test(pd,"pd_create_6_723")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_724...");
pd = pd_create_6_724();
if (!all_diagram_isotopies_test(pd,"pd_create_6_724")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_725...");
pd = pd_create_6_725();
if (!all_diagram_isotopies_test(pd,"pd_create_6_725")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_726...");
pd = pd_create_6_726();
if (!all_diagram_isotopies_test(pd,"pd_create_6_726")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_727...");
pd = pd_create_6_727();
if (!all_diagram_isotopies_test(pd,"pd_create_6_727")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_728...");
pd = pd_create_6_728();
if (!all_diagram_isotopies_test(pd,"pd_create_6_728")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_729...");
pd = pd_create_6_729();
if (!all_diagram_isotopies_test(pd,"pd_create_6_729")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_73...");
pd = pd_create_6_73();
if (!all_diagram_isotopies_test(pd,"pd_create_6_73")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_730...");
pd = pd_create_6_730();
if (!all_diagram_isotopies_test(pd,"pd_create_6_730")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_731...");
pd = pd_create_6_731();
if (!all_diagram_isotopies_test(pd,"pd_create_6_731")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_732...");
pd = pd_create_6_732();
if (!all_diagram_isotopies_test(pd,"pd_create_6_732")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_74...");
pd = pd_create_6_74();
if (!all_diagram_isotopies_test(pd,"pd_create_6_74")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_75...");
pd = pd_create_6_75();
if (!all_diagram_isotopies_test(pd,"pd_create_6_75")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_76...");
pd = pd_create_6_76();
if (!all_diagram_isotopies_test(pd,"pd_create_6_76")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_77...");
pd = pd_create_6_77();
if (!all_diagram_isotopies_test(pd,"pd_create_6_77")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_78...");
pd = pd_create_6_78();
if (!all_diagram_isotopies_test(pd,"pd_create_6_78")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_79...");
pd = pd_create_6_79();
if (!all_diagram_isotopies_test(pd,"pd_create_6_79")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_8...");
pd = pd_create_6_8();
if (!all_diagram_isotopies_test(pd,"pd_create_6_8")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_80...");
pd = pd_create_6_80();
if (!all_diagram_isotopies_test(pd,"pd_create_6_80")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_81...");
pd = pd_create_6_81();
if (!all_diagram_isotopies_test(pd,"pd_create_6_81")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_82...");
pd = pd_create_6_82();
if (!all_diagram_isotopies_test(pd,"pd_create_6_82")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_83...");
pd = pd_create_6_83();
if (!all_diagram_isotopies_test(pd,"pd_create_6_83")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_84...");
pd = pd_create_6_84();
if (!all_diagram_isotopies_test(pd,"pd_create_6_84")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_85...");
pd = pd_create_6_85();
if (!all_diagram_isotopies_test(pd,"pd_create_6_85")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_86...");
pd = pd_create_6_86();
if (!all_diagram_isotopies_test(pd,"pd_create_6_86")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_87...");
pd = pd_create_6_87();
if (!all_diagram_isotopies_test(pd,"pd_create_6_87")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_88...");
pd = pd_create_6_88();
if (!all_diagram_isotopies_test(pd,"pd_create_6_88")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_89...");
pd = pd_create_6_89();
if (!all_diagram_isotopies_test(pd,"pd_create_6_89")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_9...");
pd = pd_create_6_9();
if (!all_diagram_isotopies_test(pd,"pd_create_6_9")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_90...");
pd = pd_create_6_90();
if (!all_diagram_isotopies_test(pd,"pd_create_6_90")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_91...");
pd = pd_create_6_91();
if (!all_diagram_isotopies_test(pd,"pd_create_6_91")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_92...");
pd = pd_create_6_92();
if (!all_diagram_isotopies_test(pd,"pd_create_6_92")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_93...");
pd = pd_create_6_93();
if (!all_diagram_isotopies_test(pd,"pd_create_6_93")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_94...");
pd = pd_create_6_94();
if (!all_diagram_isotopies_test(pd,"pd_create_6_94")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_95...");
pd = pd_create_6_95();
if (!all_diagram_isotopies_test(pd,"pd_create_6_95")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_96...");
pd = pd_create_6_96();
if (!all_diagram_isotopies_test(pd,"pd_create_6_96")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_97...");
pd = pd_create_6_97();
if (!all_diagram_isotopies_test(pd,"pd_create_6_97")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_98...");
pd = pd_create_6_98();
if (!all_diagram_isotopies_test(pd,"pd_create_6_98")) {
  return false;
}
pd_code_free(&pd);

printf("testing diagram pd_create_6_99...");
pd = pd_create_6_99();
if (!all_diagram_isotopies_test(pd,"pd_create_6_99")) {
  return false;
}
pd_code_free(&pd);

return true;
}


bool test_wyes() {

  /* This is some handwritten code for testing diagram-isotopy of unknot-wyes */

  pd_code_t *wye111,
    *wye120,*wye012,*wye201,
    *wye021,*wye210,*wye102,
    *wye300,*wye030,*wye003;

  pd_code_t *group1[1];
  pd_code_t *group2a[3];
  pd_code_t *group2b[3];
  pd_code_t *group3[3];

  printf("---------------------------------------------\n"
	 "diagram_isotopy testing unknotwyes with n = 3\n"
	 "---------------------------------------------\n\n");

  printf("building all 3-crossing wyes...");

  wye111 = pd_build_unknot_wye(1,1,1);

  group1[0] = wye111;

  wye120 = pd_build_unknot_wye(1,2,0);
  wye012 = pd_build_unknot_wye(0,1,2);
  wye201 = pd_build_unknot_wye(2,0,1);
  wye021 = pd_build_unknot_wye(0,2,1);
  wye210 = pd_build_unknot_wye(2,1,0);
  wye102 = pd_build_unknot_wye(1,0,2);

  group2a[0] = wye120; group2a[1] = wye012; group2a[2] = wye201;
  group2b[0] = wye021; group2b[1] = wye210; group2b[2] = wye102;

  wye300 = pd_build_unknot_wye(3,0,0);
  wye030 = pd_build_unknot_wye(0,3,0);
  wye003 = pd_build_unknot_wye(0,0,3);

  group3[0] = wye300; group3[1] = wye030; group3[2] = wye003;

  printf("done\n");

  printf("checking different groups are NOT diagram-isotopic...");

  int i,j;

  for(i=0;i<1;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group1[i],group2a[j])) {

	printf("FAIL (%d of group 1 == %d of group 2a)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<1;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group1[i],group2b[j])) {

	printf("FAIL (%d of group 1 == %d of group 2b)\n",
	       i,j);
	return false;

      }

    }

  }
  
  for(i=0;i<1;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group1[i],group3[j])) {

	printf("FAIL (%d of group 1 == %d of group 3)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group2a[i],group3[j])) {

	printf("FAIL (%d of group 2a == %d of group 3)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group2b[i],group3[j])) {

	printf("FAIL (%d of group 2b == %d of group 3)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (pd_diagram_isotopic(group2a[i],group2b[j])) {

	printf("FAIL (%d of group 2a == %d of group 2b)\n",
	       i,j);
	return false;

      }

    }

  }

  printf("pass\n");

  printf("checking members of groups ARE isotopic to each other...");

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (!pd_diagram_isotopic(group2a[i],group2a[j])) {

	printf("FAIL (%d of group 2 != %d of group 2)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (!pd_diagram_isotopic(group2b[i],group2b[j])) {

	printf("FAIL (%d of group 2b != %d of group 2b)\n",
	       i,j);
	return false;

      }

    }

  }

  for(i=0;i<3;i++) {

    for(j=0;j<3;j++) {

      if (!pd_diagram_isotopic(group3[i],group3[j])) {

	printf("FAIL (%d of group 3 != %d of group 3)\n",
	       i,j);
	return false;

      }

    }

  }

  printf("pass\n");

  printf("housekeeping...");

  for(i=0;i<1;i++) {

    pd_code_free(&(group1[i]));

  }

  for(i=0;i<3;i++) {

    pd_code_free(&(group2a[i]));

  }

  for(i=0;i<3;i++) {

    pd_code_free(&(group2b[i]));

  }
  
  for(i=0;i<3;i++) {

    pd_code_free(&(group3[i]));

  }

  printf("done\n");

  printf("------------------------------------\n"
	 "diagram isotopy unknot wye PASS \n"
	 "------------------------------------\n");

  return true;

}      

int main() {

  printf("test_diagram_isotopic (%s)\n",PACKAGE_STRING);
  printf("---------------------------------------\n"
	 "Unit tests for pd_diagram_isotopic code\n"
	 "in pd_isomorphisms.c\n"
	 "=======================================\n");

  if (!test_wyes() || !test_diagrams() || !test_diagrams_6()) {

    printf("=====================================\n");
    printf("test_diagram_isotopic:  FAIL.\n");
    exit(1);

  } else {

    printf("=====================================\n");
    printf("test_diagram_isotopic:  PASS.\n");
    exit(0);

  }

  return 0;

}
  
  

  

  

  
